# -*- coding: utf-8 -*-
# File: test_cats.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module mapper.cats
"""

from unittest.mock import MagicMock, Mock, patch

import pytest

from deepdoctection.datapoint import CategoryAnnotation, Image, SummaryAnnotation
from deepdoctection.mapper import cat_to_sub_cat, filter_cat, filter_summary, image_to_cat_id, pub_to_image, remove_cats
from deepdoctection.utils.detection_types import JsonDict

# from deepdoctection.utils.settings import names
from deepdoctection.utils.settings import CellType, LayoutType, TableType, get_type

from .conftest import get_pubtabnet_white_image
from .data import DatapointPubtabnet


@pytest.mark.basic
@patch("deepdoctection.mapper.pubstruct.load_image_from_file", MagicMock(side_effect=get_pubtabnet_white_image))
def test_cat_to_sub_cat(datapoint_pubtabnet: JsonDict, pubtabnet_results: DatapointPubtabnet) -> None:
    """
    test func: cat_to_sub_cat replaces categories with sub categories correctly
    """
    # Arrange
    categories_name_as_key_init = {LayoutType.cell: "1", TableType.item: "2"}
    pub_to_image_mapper = pub_to_image(categories_name_as_key_init, False, False, True, False, False, False)
    dp = pub_to_image_mapper(datapoint_pubtabnet)

    categories = MagicMock()
    categories._cat_to_sub_cat = {  # pylint: disable=W0212
        LayoutType.cell: CellType.header,
        TableType.item: TableType.item,
    }
    categories.get_categories = Mock(
        return_value={CellType.header: "1", CellType.body: "2", LayoutType.row: "3", LayoutType.column: "4"}
    )

    datapoint = pubtabnet_results

    # Act
    if dp is not None:
        cat_to_sub_cat_mapper = cat_to_sub_cat(
            categories.get_categories(), categories._cat_to_sub_cat  # pylint: disable=W0212
        )
        dp = cat_to_sub_cat_mapper(dp)

    if dp is not None:
        heads = dp.get_annotation(category_names=CellType.header)
        bodies = dp.get_annotation(category_names=CellType.body)
        rows = dp.get_annotation(category_names=LayoutType.row)
        cols = dp.get_annotation(category_names=LayoutType.column)

        # Assert
        assert len(heads) == datapoint.get_number_of_heads()
        assert len(bodies) == datapoint.get_number_of_bodies()
        assert str(len(rows)) == datapoint.get_summary_ann_sub_category_rows_id()
        assert str(len(cols)) == datapoint.get_summary_ann_sub_category_col_id()


@pytest.mark.basic
@patch("deepdoctection.mapper.pubstruct.load_image_from_file", MagicMock(side_effect=get_pubtabnet_white_image))
def test_filter_categories(datapoint_pubtabnet: JsonDict, pubtabnet_results: DatapointPubtabnet) -> None:
    """
    test func:`filter_categories` removes categories correctly. Also tests that category ids are re assigned correctly
    """

    # Arrange
    categories_name_as_key_init = {LayoutType.cell: "1", TableType.item: "2"}
    pub_to_image_mapper = pub_to_image(categories_name_as_key_init, False, False, True, False, False, False)
    dp = pub_to_image_mapper(datapoint_pubtabnet)
    assert dp is not None

    categories = MagicMock()
    categories.get_categories = Mock(return_value=[TableType.item])

    datapoint = pubtabnet_results

    # Act
    filter_cat_mapper = filter_cat(  # pylint: disable=E1120
        categories.get_categories(as_dict=False, filtered=True),
        list(categories_name_as_key_init.keys()),
    )
    dp = filter_cat_mapper(dp)

    items = dp.get_annotation(category_names=TableType.item)
    cells = dp.get_annotation(category_names=LayoutType.cell)

    # Assert
    assert len(items) == int(datapoint.get_summary_ann_sub_category_rows_id()) + int(
        datapoint.get_summary_ann_sub_category_col_id()
    )
    assert items[0].category_id == "1"
    assert len(cells) == 0


@pytest.mark.basic
def test_filter_summary_1(datapoint_image_with_summary: Image) -> None:
    """
    test func:`filter_summary` does not filter dataset, if condition is satisfied.
    """

    # Arrange
    output = filter_summary({"BAK": "FOO"}, True)(datapoint_image_with_summary)

    # Assert
    assert output is not None


@pytest.mark.basic
def test_filter_summary_2(datapoint_image_with_summary: Image) -> None:
    """
    test func:`filter_summary`  does filter dataset, if condition is not satisfied.
    """

    # Arrange
    output = filter_summary({"BAK": ["BAZ"]}, True)(datapoint_image_with_summary)

    # Assert
    assert output is None


@pytest.mark.basic
def test_image_to_cat_id_1(dp_image_fully_segmented: Image) -> None:
    """
    test func: image_to_cat_id returns extraction of category_ids
    """

    # Arrange
    category_names = LayoutType.table
    expected_output = [2]

    # Act
    output, _ = image_to_cat_id(category_names)(dp_image_fully_segmented)  # pylint: disable = E1102

    # Assert
    assert output[LayoutType.table] == expected_output


@pytest.mark.basic
def test_image_to_cat_id_2(dp_image_fully_segmented: Image) -> None:
    """
    test func: image_to_cat_id returns extraction of category_ids
    """

    # Arrange
    category_names = [LayoutType.table, LayoutType.row, LayoutType.column]
    expected_output = {LayoutType.table: [2], LayoutType.row: [6, 6], LayoutType.column: [7, 7]}

    # Act
    output, _ = image_to_cat_id(category_names)(dp_image_fully_segmented)  # pylint: disable = E1102

    # Assert
    assert output[LayoutType.table] == expected_output[LayoutType.table]
    assert output[LayoutType.row] == expected_output[LayoutType.row]
    assert output[LayoutType.column] == expected_output[LayoutType.column]


@pytest.mark.basic
def test_image_to_cat_id_3(dp_image_fully_segmented: Image) -> None:
    """
    test func: image_to_cat_id returns extraction of category_ids
    """

    # Arrange
    sub_category_names = {LayoutType.cell: CellType.row_span}
    expected_output = {CellType.row_span: [1, 1, 1, 1, 0]}

    # Act
    output, _ = image_to_cat_id(sub_categories=sub_category_names)(dp_image_fully_segmented)  # pylint: disable = E1102

    # Assert
    assert output[CellType.row_span] == expected_output[CellType.row_span]


@pytest.mark.basic
def test_image_to_cat_id_4(dp_image_fully_segmented: Image) -> None:
    """
    test func: image_to_cat_id returns extraction of category_ids
    """

    # Arrange
    sub_category_names = {
        LayoutType.cell: [CellType.row_number, CellType.row_span, CellType.column_number, CellType.column_span]
    }
    expected_output = {
        CellType.row_number: [1, 2, 1, 2, 0],
        CellType.row_span: [1, 1, 1, 1, 0],
        CellType.column_number: [1, 1, 2, 2, 0],
        CellType.column_span: [1, 1, 1, 1, 0],
    }

    # Act
    output, _ = image_to_cat_id(sub_categories=sub_category_names)(dp_image_fully_segmented)  # pylint: disable = E1102

    # Assert
    assert output[CellType.row_number] == expected_output[CellType.row_number]
    assert output[CellType.row_span] == expected_output[CellType.row_span]
    assert output[CellType.column_number] == expected_output[CellType.column_number]
    assert output[CellType.column_span] == expected_output[CellType.column_span]


@pytest.mark.basic
def test_remove_cats(dp_image_fully_segmented: Image) -> None:
    """
    test func: remove_cats returns datapoint with removed categories
    """

    # Arrange
    categories = LayoutType.row

    # Act
    remove_cats_mapper = remove_cats(category_names=categories)  # pylint: disable=E1120  # 259
    dp = remove_cats_mapper(dp_image_fully_segmented)

    # Assert
    anns = dp.get_annotation(category_names=LayoutType.row)
    assert len(anns) == 0


@pytest.mark.basic
def test_remove_cats_2(dp_image_fully_segmented: Image) -> None:
    """
    test func: remove_cats returns datapoint with removed sub categories
    """

    # Arrange
    sub_categories = {LayoutType.cell: [CellType.row_number, CellType.row_span], LayoutType.row: CellType.row_number}

    # Act
    remove_cats_mapper = remove_cats(sub_categories=sub_categories)  # pylint: disable=E1120  # 259
    dp = remove_cats_mapper(dp_image_fully_segmented)

    # Assert
    anns_row = dp.get_annotation(category_names=LayoutType.row)
    anns_cell = dp.get_annotation(category_names=LayoutType.cell)

    assert len(anns_row) == 2
    assert len(anns_cell) == 5

    first_ann_row = anns_row[0]

    with pytest.raises(Exception):
        first_ann_row.get_sub_category(CellType.row_number)

    first_ann_cell = anns_cell[0]
    scd_ann_cell = anns_cell[1]

    with pytest.raises(Exception):
        first_ann_cell.get_sub_category(CellType.row_number)

    with pytest.raises(Exception):
        first_ann_cell.get_sub_category(CellType.row_span)

    assert isinstance(first_ann_cell.get_sub_category(CellType.column_number), CategoryAnnotation)
    assert isinstance(first_ann_cell.get_sub_category(CellType.column_span), CategoryAnnotation)

    with pytest.raises(Exception):
        scd_ann_cell.get_sub_category(CellType.row_number)

    with pytest.raises(Exception):
        scd_ann_cell.get_sub_category(CellType.row_span)

    assert isinstance(scd_ann_cell.get_sub_category(CellType.column_number), CategoryAnnotation)
    assert isinstance(scd_ann_cell.get_sub_category(CellType.column_span), CategoryAnnotation)


@pytest.mark.basic
def test_remove_cats_3(dp_image_fully_segmented: Image) -> None:
    """
    test func: remove_cats removes summary sub category
    """

    # Arrange
    sub_category_ann = CategoryAnnotation(category_name="TEST")
    summary = SummaryAnnotation()
    summary.dump_sub_category(get_type("TEST_SUMMARY"), sub_category_ann)
    dp_image_fully_segmented.summary = summary

    # Act
    remove_cats_mapper = remove_cats(summary_sub_categories="TEST_SUMMARY")  # pylint: disable=E1120  # 259
    dp = remove_cats_mapper(dp_image_fully_segmented)
    assert dp.summary is not None

    # Assert
    with pytest.raises(Exception):
        dp.summary.get_sub_category(get_type("TEST_SUMMARY"))



# -*- coding: utf-8 -*-
# File: xfundstruct.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for mapping annotations to and from xfund data structure
"""

import os
from collections import defaultdict
from itertools import chain
from typing import Mapping, Optional

from ..datapoint import BoundingBox, CategoryAnnotation, ContainerAnnotation, Image, ImageAnnotation
from ..utils.detection_types import JsonDict
from ..utils.fs import load_image_from_file
from ..utils.settings import (
    BioTag,
    LayoutType,
    ObjectTypes,
    Relationships,
    TokenClasses,
    WordType,
    get_type,
    token_class_tag_to_token_class_with_tag,
)
from .maputils import MappingContextManager, curry, maybe_get_fake_score


@curry
def xfund_to_image(
    dp: JsonDict,
    load_image: bool,
    fake_score: bool,
    categories_dict_name_as_key: Mapping[str, str],
    token_class_names_mapping: Mapping[str, str],
    ner_token_to_id_mapping: Mapping[ObjectTypes, Mapping[ObjectTypes, Mapping[ObjectTypes, str]]],
) -> Optional[Image]:
    """
    Map a datapoint of annotation structure as given as from xfund or funsd dataset in to an Image structure

    :param dp: A datapoint in dict structure as returned from the xfund or funsd dataset. Each datapoint must coincide
               with exactly one image sample.
    :param load_image: If 'True' it will load image to attr:`Image.image`
    :param fake_score: If dp does not contain a score, a fake score with uniform random variables in (0,1)
                       will be added.
    :param categories_dict_name_as_key:
    :param token_class_names_mapping: A dictionary, mapping original label names to normalized category names
    :param ner_token_to_id_mapping: A dictionary, mapping token classes with bio tags (i.e. token tags) into their
                                    category ids.
    :return: Image
    """

    img = dp.get("img")
    if img is None:
        full_path = dp.get("file_name")
    else:
        full_path = img.get("fname")

    if full_path is None:
        return None

    _, file_name = os.path.split(full_path)
    external_id = dp.get("uid")
    tag_to_id_mapping = ner_token_to_id_mapping[LayoutType.word][WordType.tag]
    token_class_to_id_mapping = ner_token_to_id_mapping[LayoutType.word][WordType.token_class]
    token_tag_to_id_mapping = ner_token_to_id_mapping[LayoutType.word][WordType.token_tag]

    with MappingContextManager(file_name) as mapping_context:

        image = Image(file_name=file_name, location=full_path, external_id=external_id)

        image.image = load_image_from_file(full_path)

        if not load_image:
            image.clear_image()

        entity_id_to_ann_id = defaultdict(list)
        entity_id_to_entity_link_id = defaultdict(list)
        ann_id_to_entity_id = {}

        entities = dp.get("document", [])

        if not entities:
            entities = dp.get("form", [])

        for entity in entities:
            box = list(map(float, entity["box"]))
            bbox = BoundingBox(absolute_coords=True, ulx=box[0], uly=box[1], lrx=box[2], lry=box[3])
            score = maybe_get_fake_score(fake_score)
            entity_ann = ImageAnnotation(
                category_name=LayoutType.text,
                bounding_box=bbox,
                category_id=categories_dict_name_as_key[LayoutType.text],
                score=score,
            )
            category_name = token_class_names_mapping[entity["label"]]
            sub_cat_semantic = CategoryAnnotation(
                category_name=category_name, category_id=token_class_to_id_mapping[get_type(category_name)]
            )
            entity_ann.dump_sub_category(WordType.token_class, sub_cat_semantic)
            image.dump(entity_ann)

            words = entity.get("words")

            for idx, word in enumerate(words):
                box = list(map(float, word["box"]))
                bbox = BoundingBox(absolute_coords=True, ulx=box[0], uly=box[1], lrx=box[2], lry=box[3])

                score = maybe_get_fake_score(fake_score)

                ann = ImageAnnotation(
                    category_name=LayoutType.word,
                    bounding_box=bbox,
                    category_id=categories_dict_name_as_key[LayoutType.word],
                    score=score,
                )
                image.dump(ann)
                entity_ann.dump_relationship(Relationships.child, ann.annotation_id)
                sub_cat_semantic = CategoryAnnotation(
                    category_name=category_name, category_id=token_class_to_id_mapping[get_type(category_name)]
                )
                ann.dump_sub_category(WordType.token_class, sub_cat_semantic)
                sub_cat_chars = ContainerAnnotation(category_name=WordType.characters, value=word["text"])
                ann.dump_sub_category(WordType.characters, sub_cat_chars)
                if sub_cat_semantic.category_name == TokenClasses.other:
                    sub_cat_tag = CategoryAnnotation(
                        category_name=BioTag.outside, category_id=tag_to_id_mapping[BioTag.outside]
                    )
                    ann.dump_sub_category(WordType.tag, sub_cat_tag)
                    # populating ner token to be used for training and evaluation
                    sub_cat_ner_tok = CategoryAnnotation(
                        category_name=BioTag.outside, category_id=token_tag_to_id_mapping[BioTag.outside]
                    )
                    ann.dump_sub_category(WordType.token_tag, sub_cat_ner_tok)
                elif not idx:
                    sub_cat_tag = CategoryAnnotation(
                        category_name=BioTag.begin, category_id=tag_to_id_mapping[BioTag.begin]
                    )
                    ann.dump_sub_category(WordType.tag, sub_cat_tag)
                    sub_cat_ner_tok = CategoryAnnotation(
                        category_name=token_class_tag_to_token_class_with_tag(
                            get_type(sub_cat_semantic.category_name), BioTag.begin
                        ),
                        category_id=token_tag_to_id_mapping[
                            token_class_tag_to_token_class_with_tag(
                                get_type(sub_cat_semantic.category_name), BioTag.begin
                            )
                        ],
                    )
                    ann.dump_sub_category(WordType.token_tag, sub_cat_ner_tok)
                else:
                    sub_cat_tag = CategoryAnnotation(
                        category_name=BioTag.inside, category_id=tag_to_id_mapping[BioTag.inside]
                    )
                    ann.dump_sub_category(WordType.tag, sub_cat_tag)
                    sub_cat_ner_tok = CategoryAnnotation(
                        category_name=token_class_tag_to_token_class_with_tag(
                            get_type(sub_cat_semantic.category_name), BioTag.inside
                        ),
                        category_id=token_tag_to_id_mapping[
                            token_class_tag_to_token_class_with_tag(
                                get_type(sub_cat_semantic.category_name), BioTag.inside
                            )
                        ],
                    )
                    ann.dump_sub_category(WordType.token_tag, sub_cat_ner_tok)

                entity_id_to_ann_id[entity["id"]].append(ann.annotation_id)
                ann_id_to_entity_id[ann.annotation_id] = entity["id"]

            entity_id_to_entity_link_id[entity["id"]].extend(entity["linking"])

        # now populating semantic links
        word_anns = image.get_annotation(category_names=LayoutType.word)
        for word in word_anns:
            entity_id = ann_id_to_entity_id[word.annotation_id]
            all_linked_entities = list(chain(*entity_id_to_entity_link_id[entity_id]))
            ann_ids = []
            for linked_entity in all_linked_entities:
                ann_ids.extend(entity_id_to_ann_id[linked_entity])
            for ann_id in ann_ids:
                if ann_id != word.annotation_id:
                    word.dump_relationship(Relationships.semantic_entity_link, ann_id)

    if mapping_context.context_error:
        return None
    return image



# -*- coding: utf-8 -*-
# File: layouttest.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for Testlayout dataset. Install the dataset following the folder structure

    testlayout
    ├── predict
    │ ├── xrf_layout_test_predict.jsonl
    ├── test
    │ ├── xrf_layout_test.jsonl
"""

from typing import Mapping, Union

from ...dataflow import DataFlow, MapData
from ...dataflow.custom_serialize import SerializerJsonlines
from ...datasets.info import DatasetInfo
from ...mapper.prodigystruct import prodigy_to_image
from ...utils.settings import DatasetType, LayoutType
from ..base import _BuiltInDataset
from ..dataflow_builder import DataFlowBaseBuilder
from ..info import DatasetCategories
from ..registry import dataset_registry

_NAME = "testlayout"
_DESCRIPTION = (
    "A small test set consisting of a subset of six images with layout annotations. \n"
    "The set is meant for debugging purposes and nothing else"
)

_LICENSE = (
    "The annotations in this dataset belong to Dr. Janis Meyer and are licensed \n"
    " under a Community Data License Agreement \n"
    " – Permissive – Version 1.0 License. Dr. Janis Meyer does not own the copyright of the images. \n"
    " Use of the images must abide by the PMC Open Access Subset Terms of Use."
)
_URL = [
    "https://www.googleapis.com/drive/v3/files/1ZD4Ef4gd2FIfp7vR8jbnrZeXD3gSWNqE?alt"
    "=media&key=AIzaSyDuoPG6naK-kRJikScR7cP_1sQBF1r3fWU",
    "https://www.googleapis.com/drive/v3/files/18HD62LFLa1iAmqffo4SyjuEQ32MzyNQ0?alt"
    "=media&key=AIzaSyDuoPG6naK-kRJikScR7cP_1sQBF1r3fWU",
]
_SPLITS: Mapping[str, str] = {"test": "test", "predict": "predict"}
_TYPE = DatasetType.object_detection
_LOCATION = "testlayout"

_ANNOTATION_FILES: Mapping[str, str] = {
    "test": "xrf_layout_test.jsonl",
    "predict": "xrf_layout_test_predict.jsonl",
}

_INIT_CATEGORIES = [LayoutType.text, LayoutType.title, LayoutType.list, LayoutType.table, LayoutType.figure]


@dataset_registry.register("testlayout")
class LayoutTest(_BuiltInDataset):
    """
    LayoutTest
    """

    _name = _NAME

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name=_NAME, description=_DESCRIPTION, license=_LICENSE, url=_URL, splits=_SPLITS, type=_TYPE)

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES)

    def _builder(self) -> "LayoutTestBuilder":
        return LayoutTestBuilder(location=_LOCATION, annotation_files=_ANNOTATION_FILES)


class LayoutTestBuilder(DataFlowBaseBuilder):
    """
    LayoutTest dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the returns
        of the dataflow:

        `split:` Split of the dataset. Only "test" is for this small sample available

        `max_datapoints:` Will stop iterating after max_datapoints. Default: `None`

        `load_image:` Will load the image for each datapoint.  Default: `False`

        `fake_score:` Will add a fake score so that annotations look like predictions

        :return: Dataflow
        """
        split = str(kwargs.get("split", "test"))
        max_datapoints = kwargs.get("max_datapoints")
        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)
        load_image = kwargs.get("load_image", False)
        fake_score = kwargs.get("fake_score", False)

        # Load
        df: DataFlow
        path = self.get_workdir() / self.splits[split] / self.get_annotation_file(split)

        df = SerializerJsonlines.load(path, max_datapoints=max_datapoints)

        # Map
        df = MapData(df, lambda dp: dp if dp["answer"] == "accept" else None)
        prodigy_mapper = prodigy_to_image(  # pylint: disable=E1120
            categories_name_as_key=self.categories.get_categories(init=True, name_as_key=True),
            load_image=load_image,
            fake_score=fake_score,
        )
        df = MapData(df, prodigy_mapper)

        return df



# -*- coding: utf-8 -*-
# File: conftest.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Fixture module for train
"""
from copy import deepcopy
from dataclasses import dataclass
from typing import Union

import numpy as np
from pytest import fixture

from deepdoctection.dataflow import DataFlow, DataFromList
from deepdoctection.datapoint import BoundingBox, Image, ImageAnnotation
from deepdoctection.datasets.base import DatasetBase
from deepdoctection.datasets.dataflow_builder import DataFlowBaseBuilder
from deepdoctection.datasets.info import DatasetCategories, DatasetInfo
from deepdoctection.utils.settings import LayoutType

_NAME = "test_set"
_LOCATION = "/path/to/dir"
_INIT_CATEGORIES = [LayoutType.table, LayoutType.text]
_DESCRIPTION = "test description"
_URL = "test url"
_LICENSE = "test license"


@dataclass
class Datapoint:
    """
    A dataclass for generating an Image datapoint with annotations
    """

    image = Image(location="/test/to/path", file_name="test_name")
    np_image = np.ones([400, 600, 3], dtype=np.uint8)
    anns = [
        ImageAnnotation(
            bounding_box=BoundingBox(ulx=100.0, uly=100.0, lrx=200.0, lry=400.0, absolute_coords=True),
            category_name="table",
            category_id="1",
        ),
        ImageAnnotation(
            bounding_box=BoundingBox(ulx=50.0, uly=70.0, lrx=70.0, lry=90.0, absolute_coords=True),
            category_name="text",
            category_id="2",
        ),
    ]

    def get_datapoint(self) -> Image:
        """
        datapoint
        """
        image = deepcopy(self.image)
        image.image = self.np_image
        for ann in self.anns:
            image.dump(ann)
        return image


class TestDataset(DatasetBase):
    """
    Class for a dataset fixture
    """

    @classmethod
    def _info(cls) -> DatasetInfo:
        """
        _info
        """
        return DatasetInfo(name=_NAME, description=_DESCRIPTION, url=_URL, license=_LICENSE)

    def _categories(self) -> DatasetCategories:
        """
        _categories
        """
        return DatasetCategories(init_categories=_INIT_CATEGORIES)

    def _builder(self) -> "TestBuilder":
        """
        _builder
        """
        return TestBuilder(location=_LOCATION)


class TestBuilder(DataFlowBaseBuilder):
    """
    test dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        build
        """
        return DataFromList([Datapoint().get_datapoint(), Datapoint().get_datapoint()])


@fixture(name="test_dataset")
def fixture_test_dataset() -> DatasetBase:
    """
    fixture for test dataset
    """
    return TestDataset()



# -*- coding: utf-8 -*-
# File: tpcompat.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Compatibility classes and methods related to Tensorpack package
"""

from abc import ABC, abstractmethod
from typing import Any, List, Mapping, Tuple, Union

from tensorpack.predict import OfflinePredictor, PredictConfig  # pylint: disable=E0401
from tensorpack.tfutils import SmartInit  # pylint: disable=E0401

# pylint: disable=import-error
from tensorpack.train.model_desc import ModelDesc
from tensorpack.utils.gpu import get_num_gpu

from ...utils.metacfg import AttrDict
from ...utils.settings import ObjectTypes

# pylint: enable=import-error


class ModelDescWithConfig(ModelDesc, ABC):  # type: ignore
    """
    A wrapper for Tensorpack ModelDesc for bridging the gap between Tensorpack and DD API. Only for storing a
    configuration of hyperparameters and maybe training settings.
    """

    def __init__(self, config: AttrDict) -> None:
        """
        :param config: Config setting
        """
        super().__init__()
        self.cfg = config

    def get_inference_tensor_names(self) -> Tuple[List[str], List[str]]:
        """
        Returns lists of tensor names to be used to create an inference callable. "build_graph" must create tensors
        of these names when called under inference context.

        :return: Tuple of list input and list output names. The names must coincide with tensor within the model.
        """
        raise NotImplementedError


class TensorpackPredictor(ABC):
    """
    The base class for wrapping a Tensorpack predictor. It takes a ModelDescWithConfig from Tensorpack and weights and
    builds a Tensorpack offline predictor (e.g. a default session will be generated when initializing).

    Two abstract methods need to be implemented:

        - `set_model` generates a ModelDescWithConfig from the input .yaml file and possible manual adaptions of
          the configuration.

        - `predict` the interface of this class for calling the OfflinePredictor and returning the results. This
          method will be used throughout in the context of pipeline components. If there are some pre- or
          post-processing steps, you can place them here. However, do not convert the returned results into DD objects
          as there is an explicit class available for this.
    """

    def __init__(self, model: ModelDescWithConfig, path_weights: str, ignore_mismatch: bool) -> None:
        """
        :param model: Model, either as ModelDescWithConfig or derived from that class.
        :param path_weights: Model weights of the prediction config.
        :param ignore_mismatch: When True will ignore mismatches between checkpoint weights and models. This is needed
                                if a pre-trained model is to be fine-tuned on a custom dataset.
        """
        self._model = model
        self.path_weights = path_weights
        self.ignore_mismatch = ignore_mismatch
        self._number_gpus = get_num_gpu()
        self.predict_config = self._build_config()
        self.tp_predictor = self.get_predictor()

    def get_predictor(self) -> OfflinePredictor:
        """
        :return: Returns an OfflinePredictor.
        """
        return OfflinePredictor(self.predict_config)

    def _build_config(self) -> PredictConfig:
        predict_config = PredictConfig(
            model=self._model,
            session_init=SmartInit(self.path_weights, ignore_mismatch=self.ignore_mismatch),
            input_names=self._model.get_inference_tensor_names()[0],
            output_names=self._model.get_inference_tensor_names()[1],
        )

        return predict_config

    @staticmethod
    @abstractmethod
    def set_model(
        path_yaml: str, categories: Mapping[str, ObjectTypes], config_overwrite: Union[List[str], None]
    ) -> ModelDescWithConfig:
        """
        Implement the config generation, its modification and instantiate a version of the model. See
        `pipe.tpfrcnn.TPFrcnnDetector` for an example
        """
        raise NotImplementedError

    @abstractmethod
    def predict(self, np_img: Any) -> Any:
        """
        Implement, how `self.tp_predictor` is invoked and raw prediction results are generated. Do use only raw
        objects and nothing, which is related to the DD API.
        """
        raise NotImplementedError

    @property
    def model(self) -> ModelDescWithConfig:
        """
        model
        """
        return self._model



# -*- coding: utf-8 -*-
# File: viz.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some visualisation utils. Copied and pasted from

<https://github.com/tensorpack/tensorpack/blob/master/tensorpack/utils/viz.py>

and

<https://github.com/facebookresearch/detectron2/blob/main/detectron2/utils/colormap.py>
"""
import sys
from typing import Callable, List, Optional, Tuple

import cv2
import numpy as np
import numpy.typing as npt
from numpy import float32

from .detection_types import ImageType

__all__ = ["draw_text", "draw_boxes", "interactive_imshow"]

_COLORS = (
    np.array(
        [
            0.667,
            0.333,
            0.500,
            0.667,
            0.667,
            0.500,
            0.667,
            1.000,
            0.500,
            1.000,
            0.000,
            0.500,
            1.000,
            0.333,
            0.500,
            1.000,
            0.667,
            0.500,
            1.000,
            1.000,
            0.500,
            0.000,
            0.333,
            1.000,
            0.000,
            0.667,
            1.000,
            0.000,
            1.000,
            1.000,
            0.333,
            0.000,
            1.000,
            0.333,
            0.333,
            1.000,
            0.333,
            0.667,
            1.000,
            0.333,
            1.000,
            1.000,
            0.667,
            0.000,
            1.000,
            0.667,
            0.333,
            1.000,
            0.667,
            0.667,
            1.000,
            0.667,
            1.000,
            1.000,
            1.000,
            0.000,
            1.000,
            1.000,
            0.333,
            1.000,
            1.000,
            0.667,
            1.000,
            0.333,
            0.000,
            0.000,
            0.500,
            0.000,
            0.000,
            0.667,
            0.000,
            0.000,
            0.833,
            0.000,
            0.000,
            1.000,
            0.000,
            0.000,
            0.000,
            0.167,
            0.000,
            0.000,
            0.333,
            0.000,
            0.000,
            0.500,
            0.000,
            0.000,
            0.667,
            0.000,
            0.000,
            0.833,
            0.000,
            0.000,
            1.000,
            0.000,
            0.000,
            0.000,
            0.167,
            0.000,
            0.000,
            0.333,
            0.000,
            0.000,
            0.500,
            0.000,
            0.000,
            0.667,
            0.000,
            0.000,
            0.833,
            0.000,
            0.000,
            1.000,
            0.000,
            0.000,
            0.000,
            0.143,
            0.143,
            0.143,
            0.857,
            0.857,
            0.857,
        ]
    )
    .astype(np.float32)
    .reshape(-1, 3)
)


def random_color(rgb: bool = True, maximum: int = 255) -> Tuple[int, int, int]:
    """
    :param rgb: Whether to return RGB colors or BGR colors.
    :param maximum: either 255 or 1
    :return:
    """

    idx = np.random.randint(0, len(_COLORS))
    ret = _COLORS[idx] * maximum
    if not rgb:
        ret = ret[::-1]
    return tuple(int(x) for x in ret)  # type: ignore


def draw_text(
    np_image: ImageType, pos: Tuple[int, int], text: str, color: Tuple[int, int, int], font_scale: float = 1.0
) -> ImageType:
    """
    Draw text on an image.

    :param np_image: image as np.ndarray
    :param pos: x, y; the position of the text
    :param text: text string to draw
    :param color: a 3-tuple BGR color in [0, 255]
    :param font_scale: float
    :return: numpy array
    """

    np_image = np_image.astype(np.uint8)
    x_0, y_0 = int(pos[0]), int(pos[1])
    # Compute text size.
    font = cv2.FONT_HERSHEY_SIMPLEX
    ((text_w, text_h), _) = cv2.getTextSize(text, font, font_scale, 1)
    # Place text background.
    if x_0 + text_w > np_image.shape[1]:
        x_0 = np_image.shape[1] - text_w
    if y_0 - int(1.15 * text_h) < 0:
        y_0 = int(1.15 * text_h)
    back_top_left = x_0, y_0 - int(1.3 * text_h)
    back_bottom_right = x_0 + text_w, y_0
    cv2.rectangle(np_image, back_top_left, back_bottom_right, color, -1)
    # Show text.
    text_bottomleft = x_0, y_0 - int(0.25 * text_h)
    cv2.putText(np_image, text, text_bottomleft, font, font_scale, (0, 0, 0), thickness=1, lineType=cv2.LINE_AA)
    return np_image


def draw_boxes(
    np_image: ImageType,
    boxes: npt.NDArray[float32],
    category_names_list: Optional[List[Optional[str]]] = None,
    color: Optional[Tuple[int, int, int]] = None,
    font_scale: float = 1.0,
    rectangle_thickness: int = 4,
) -> ImageType:
    """
    Dray bounding boxes with category names into image.

    :param np_image: Image as np.ndarray
    :param boxes: A numpy array of shape Nx4 where each row is [x1, y1, x2, y2].
    :param category_names_list: List of N category names.
    :param color: A 3-tuple BGR color (in range [0, 255])
    :param font_scale: Font scale of text box
    :param rectangle_thickness: Thickness of bounding box
    :return: A new image np.ndarray
    """

    boxes = np.asarray(boxes, dtype="int32")
    if category_names_list is not None:
        assert len(category_names_list) == len(boxes), f"{len(category_names_list)} != {len(boxes)}"
    areas = (boxes[:, 2] - boxes[:, 0] + 1) * (boxes[:, 3] - boxes[:, 1] + 1)
    sorted_inds = np.argsort(-areas)  # draw large ones first
    assert areas.min() > 0, areas.min()
    # allow equal, because we are not very strict about rounding error here
    assert (
        boxes[:, 0].min() >= 0
        and boxes[:, 1].min() >= 0
        and boxes[:, 2].max() <= np_image.shape[1]
        and boxes[:, 3].max() <= np_image.shape[0]
    ), f"Image shape: {str(np_image.shape)}\n Boxes:\n{str(boxes)}"

    np_image = np_image.copy()

    if np_image.ndim == 2 or (np_image.ndim == 3 and np_image.shape[2] == 1):
        np_image = cv2.cvtColor(np_image, cv2.COLOR_GRAY2BGR)
    for i in sorted_inds:
        box = boxes[i, :]
        if category_names_list is not None:
            choose_color = random_color() if color is None else color
            if category_names_list[i] is not None:
                np_image = draw_text(
                    np_image, (box[0], box[1]), category_names_list[i], color=choose_color, font_scale=font_scale
                )
            cv2.rectangle(
                np_image, (box[0], box[1]), (box[2], box[3]), color=choose_color, thickness=rectangle_thickness
            )
    return np_image


def interactive_imshow(
    img: ImageType,
    lclick_cb: Optional[Callable[[npt.NDArray[float32], int, int], None]] = None,
    rclick_cb: Optional[Callable[[npt.NDArray[float32], int, int], None]] = None,
    **kwargs: str,
) -> None:
    """
    Display an image in a pop-up window

    :param img: An image (expect BGR) to show.
    :param lclick_cb: a callback ``func(img, x, y)`` for left/right click event.
    :param rclick_cb: a callback ``func(img, x, y)`` for left/right click event.
    :param kwargs: can be {key_cb_a: callback_img, key_cb_b: callback_img}, to specify a callback ``func(img)`` for
                   keypress. Some existing keypress event handler:

                          * q: destroy the current window
                          * x: execute ``sys.exit()``
                          * s: save image to "out.png"
    """
    name = "q, x: quit / s: save"
    cv2.imshow(name, img)

    def mouse_cb(event, x, y, *args):  # type: ignore
        if event == cv2.EVENT_LBUTTONUP and lclick_cb is not None:
            lclick_cb(img.astype(dtype=float32), x, y)
        elif event == cv2.EVENT_RBUTTONUP and rclick_cb is not None:
            rclick_cb(img.astype(dtype=float32), x, y)

    cv2.setMouseCallback(name, mouse_cb)
    key = cv2.waitKey(-1)
    while key >= 128:
        key = cv2.waitKey(-1)
    key = chr(key & 0xFF)
    cb_name = "key_cb_" + key
    if cb_name in kwargs:
        kwargs[cb_name](img)  # type: ignore
    elif key == "q":
        cv2.destroyWindow(name)
    elif key == "x":
        sys.exit()
    elif key == "s":
        cv2.imwrite("out.png", img)
    elif key in ["+", "="]:
        img = cv2.resize(img, None, fx=1.3, fy=1.3, interpolation=cv2.INTER_CUBIC)
        interactive_imshow(img, lclick_cb, rclick_cb, **kwargs)
    elif key == "-":
        img = cv2.resize(img, None, fx=0.7, fy=0.7, interpolation=cv2.INTER_CUBIC)
        interactive_imshow(img, lclick_cb, rclick_cb, **kwargs)



# -*- coding: utf-8 -*-
# File: test_iiitar13k.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module mapper.iiitarstruct
"""

from math import isclose
from typing import Dict
from unittest.mock import MagicMock, patch

from pytest import mark

from deepdoctection.mapper.pascalstruct import pascal_voc_dict_to_image
from deepdoctection.utils.detection_types import JsonDict

from .data import IIITar13KJson


@mark.basic
@patch("deepdoctection.mapper.pascalstruct.load_image_from_file", MagicMock())
def test_pascal_voc_dict_to_image(
    datapoint_iiitar13kjson: JsonDict,
    iiitar13k_categories_name_as_keys: Dict[str, str],
    iiitar13k_category_names_mapping: Dict[str, str],
    iiitar13k_results: IIITar13KJson,
) -> None:

    """
    testing iiitar13k_to_image is mapping correctly
    """

    # Act
    iiitar13k_to_image_mapper = pascal_voc_dict_to_image(
        iiitar13k_categories_name_as_keys, False, False, False, iiitar13k_category_names_mapping
    )
    dp = iiitar13k_to_image_mapper(datapoint_iiitar13kjson)

    # Assert
    datapoint = iiitar13k_results
    assert dp is not None
    test_anns = dp.get_annotation()
    assert len(test_anns) == datapoint.get_number_anns()
    assert dp.width == datapoint.get_width()
    assert dp.height == datapoint.get_height()
    assert test_anns[0].category_name == datapoint.get_first_ann_category_name()
    assert test_anns[0].bounding_box
    assert isclose(test_anns[0].bounding_box.ulx, datapoint.get_first_ann_box().ulx, rel_tol=1e-15)
    assert isclose(test_anns[0].bounding_box.uly, datapoint.get_first_ann_box().uly, rel_tol=1e-15)
    assert isclose(test_anns[0].bounding_box.width, datapoint.get_first_ann_box().w, rel_tol=1e-15)
    assert isclose(test_anns[0].bounding_box.height, datapoint.get_first_ann_box().h, rel_tol=1e-15)



# -*- coding: utf-8 -*-
# File: d2_frcnn_train.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for training Detectron2 `GeneralizedRCNN`
"""


import copy
from typing import Any, Dict, List, Mapping, Optional, Sequence, Type, Union

from detectron2.config import CfgNode, get_cfg
from detectron2.data import DatasetMapper, build_detection_train_loader
from detectron2.data.transforms import RandomFlip, ResizeShortestEdge
from detectron2.engine import DefaultTrainer, HookBase, default_writers, hooks
from detectron2.utils import comm
from detectron2.utils.events import EventWriter, get_event_storage
from fvcore.nn.precise_bn import get_bn_modules  # type: ignore
from torch.utils.data import DataLoader, IterableDataset

from ..datasets.adapter import DatasetAdapter
from ..datasets.base import DatasetBase
from ..datasets.registry import get_dataset
from ..eval.base import MetricBase
from ..eval.eval import Evaluator
from ..eval.registry import metric_registry
from ..extern.d2detect import D2FrcnnDetector
from ..extern.pt.ptutils import get_num_gpu
from ..mapper.d2struct import image_to_d2_frcnn_training
from ..pipe.base import PredictorPipelineComponent
from ..pipe.registry import pipeline_component_registry
from ..utils.file_utils import get_wandb_requirement, wandb_available
from ..utils.logger import logger
from ..utils.utils import string_to_dict

if wandb_available():
    import wandb


def _set_config(
    path_config_yaml: str,
    conf_list: List[str],
    dataset_train: DatasetBase,
    dataset_val: Optional[DatasetBase],
    metric_name: Optional[str],
    metric: Optional[Union[Type[MetricBase], MetricBase]],
    pipeline_component_name: Optional[str],
) -> CfgNode:
    cfg = get_cfg()
    cfg.defrost()
    cfg.NMS_THRESH_CLASS_AGNOSTIC = 0.01
    cfg.DATASETS.TRAIN = (dataset_train.dataset_info.name,)
    cfg.WANDB = CfgNode()
    cfg.WANDB.USE_WANDB = False
    cfg.WANDB.PROJECT = None
    cfg.WANDB.REPO = "deepdoctection"
    cfg.merge_from_file(path_config_yaml)
    cfg.merge_from_list(conf_list)

    cfg.TEST.DO_EVAL = (
        cfg.TEST.EVAL_PERIOD > 0
        and dataset_val is not None
        and (metric_name is not None or metric is not None)
        and pipeline_component_name is not None
    )
    if cfg.TEST.DO_EVAL:
        cfg.DATASETS.TEST = (dataset_val.dataset_info.name,)  # type: ignore
    cfg.freeze()
    return cfg


def _update_for_eval(config_overwrite: List[str]) -> List[str]:
    ret = [item for item in config_overwrite if not "WANDB" in item]
    return ret


class WandbWriter(EventWriter):
    """
    Write all scalars to a wandb tool.
    """

    def __init__(
        self,
        project: str,
        repo: str,
        config: Optional[Union[Dict[str, Any], CfgNode]] = None,
        window_size: int = 20,
        **kwargs: Any,
    ):
        """
        :param project: W&B Project name
        :param config: the project level configuration object
        :param window_size: the scalars will be median-smoothed by this window size
        :param kwargs: other arguments passed to `wandb.init(...)`
        """
        if config is None:
            config = {}
        self._window_size = window_size
        self._run = wandb.init(project=project, config=config, **kwargs) if not wandb.run else wandb.run
        self._run._label(repo=repo)  # type:ignore

    def write(self) -> None:
        storage = get_event_storage()

        log_dict = {}
        for key, (val, _) in storage.latest_with_smoothing_hint(self._window_size).items():
            log_dict[key] = val

        self._run.log(log_dict)  # type:ignore

    def close(self) -> None:
        self._run.finish()  # type:ignore


class D2Trainer(DefaultTrainer):
    """
    Detectron2 `DefaultTrainer` with some custom method for handling datasets and running evaluation. The setting is
    made to train standard models in detectron2.
    """

    def __init__(self, cfg: CfgNode, torch_dataset: IterableDataset[Any], mapper: DatasetMapper) -> None:
        self.dataset = torch_dataset
        self.mapper = mapper
        self.evaluator: Optional[Evaluator] = None
        self.build_val_dict: Mapping[str, str] = {}
        super().__init__(cfg)

    def build_hooks(self) -> List[HookBase]:
        """
        Overwritten from DefaultTrainer. This ensures that the EvalHook is being called before the writer and
        all metrics are being written to JSON, Tensorboard etc.

        :return: list[HookBase]
        """
        cfg = self.cfg.clone()
        cfg.defrost()
        cfg.DATALOADER.NUM_WORKERS = 0  # save some memory and time for PreciseBN

        ret = [
            hooks.IterationTimer(),
            hooks.LRScheduler(),
            hooks.PreciseBN(
                # Run at the same freq as (but before) evaluation.
                cfg.TEST.EVAL_PERIOD,
                self.model,  # pylint: disable=E1101
                # Build a new data loader to not affect training
                self.build_train_loader(cfg),
                cfg.TEST.PRECISE_BN.NUM_ITER,
            )
            if cfg.TEST.PRECISE_BN.ENABLED and get_bn_modules(self.model)  # pylint: disable=E1101
            else None,
        ]

        # Do PreciseBN before checkpointer, because it updates the model and need to
        # be saved by checkpointer.
        # This is not always the best: if checkpointing has a different frequency,
        # some checkpoints may have more precise statistics than others.
        if comm.is_main_process():
            ret.append(hooks.PeriodicCheckpointer(self.checkpointer, cfg.SOLVER.CHECKPOINT_PERIOD))

        # Do evaluation after checkpointer, because then if it fails,
        # we can use the saved checkpoint to debug.
        if self.cfg.TEST.DO_EVAL:
            ret.append(
                hooks.EvalHook(
                    cfg.TEST.EVAL_PERIOD,
                    lambda: self.eval_with_dd_evaluator(**self.build_val_dict),  # pylint: disable=W0108
                )
            )

        if comm.is_main_process():
            # Here the default print/log frequency of each writer is used.
            # run writers in the end, so that evaluation metrics are written
            ret.append(hooks.PeriodicWriter(self.build_writers(), period=20))

        return ret

    def build_writers(self) -> List[EventWriter]:
        """
        Build a list of writers to be using `default_writers()`.
        If you'd like a different list of writers, you can overwrite it in
        your trainer.

        :return: A list of `EventWriter` objects.
        """
        writers_list = default_writers(self.cfg.OUTPUT_DIR, self.max_iter)
        if self.cfg.WANDB.USE_WANDB:
            _, _wandb_available, err_msg = get_wandb_requirement()
            if not _wandb_available:
                raise ImportError(err_msg)
            if self.cfg.WANDB.PROJECT is None:
                raise ValueError("When using W&B, you must specify a project, i.e. WANDB.PROJECT")
            writers_list.append(WandbWriter(self.cfg.WANDB.PROJECT, self.cfg.WANDB.REPO, self.cfg))
        return writers_list

    def build_train_loader(self, cfg: CfgNode) -> DataLoader[Any]:  # pylint: disable=W0221
        """
        Overwritten method from `DefaultTrainer`.

        :param cfg: Configuration
        :return: The data loader for a given dataset adapter, mapper.
        """
        return build_detection_train_loader(
            dataset=self.dataset, mapper=self.mapper, total_batch_size=cfg.SOLVER.IMS_PER_BATCH
        )

    def eval_with_dd_evaluator(self, **build_eval_kwargs: str) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """
        Running the Evaluator. This method will be called from the `EvalHook`

        :param build_eval_kwargs: dataflow eval config kwargs of the underlying dataset
        :return: A dict of evaluation results
        """
        assert self.evaluator is not None
        assert self.evaluator.pipe_component is not None
        for comp in self.evaluator.pipe_component.pipe_components:
            comp.predictor.d2_predictor = copy.deepcopy(self.model).eval()  # type: ignore # pylint: disable=E1101
        scores = self.evaluator.run(True, **build_eval_kwargs)
        return scores

    def setup_evaluator(
        self,
        dataset_val: DatasetBase,
        pipeline_component: PredictorPipelineComponent,
        metric: Union[Type[MetricBase], MetricBase],
        build_val_dict: Optional[Mapping[str, str]] = None,
    ) -> None:
        """
        Setup of evaluator before starting training. During training, predictors will be replaced by current
        checkpoints.

        :param dataset_val: dataset on which to run evaluation
        :param pipeline_component: pipeline component to plug into the evaluator
        :param metric: A metric class
        :param build_val_dict: evaluation dataflow build config
        """
        self.evaluator = Evaluator(
            dataset_val,
            pipeline_component,
            metric,
            num_threads=get_num_gpu() * 2,
            run=wandb.run if wandb.run is not None else None,
        )
        if build_val_dict:
            self.build_val_dict = build_val_dict
        assert self.evaluator.pipe_component
        for comp in self.evaluator.pipe_component.pipe_components:
            assert isinstance(comp, PredictorPipelineComponent)
            assert isinstance(comp.predictor, D2FrcnnDetector)
            comp.predictor.d2_predictor = None

    @classmethod
    def build_evaluator(cls, cfg, dataset_name):  # type: ignore
        raise NotImplementedError


def train_d2_faster_rcnn(
    path_config_yaml: str,
    dataset_train: Union[str, DatasetBase],
    path_weights: str,
    config_overwrite: Optional[List[str]] = None,
    log_dir: str = "train_log/frcnn",
    build_train_config: Optional[Sequence[str]] = None,
    dataset_val: Optional[DatasetBase] = None,
    build_val_config: Optional[Sequence[str]] = None,
    metric_name: Optional[str] = None,
    metric: Optional[Union[Type[MetricBase], MetricBase]] = None,
    pipeline_component_name: Optional[str] = None,
) -> None:
    """
    Adaptation of <https://github.com/facebookresearch/detectron2/blob/main/tools/train_net.py> for training Detectron2
    standard models

    Train Detectron2 from scratch or fine-tune a model using this API. Compared to Tensorpack this framework trains much
    faster, e.g. <https://detectron2.readthedocs.io/en/latest/notes/benchmarks.html> .

    This training script is devoted to the case where one cluster with one GPU is available. To run on several machines
    with more than one GPU use `detectron2.engine.launch` .

        if __name__ == "__main__":

                launch(train_d2_faster_rcnn,
                       num_gpus,
                       num_machines,
                       machine_rank,
                       dist_url,
                       args=(path_config_yaml,
                             path_weights,
                             config_overwrite,
                             log_dir,
                             build_train_config,
                             dataset_val,
                             build_val_config,
                             metric_name,
                             metric,
                             pipeline_component_name),)


    :param path_config_yaml: path to a D2 config file. Check
                             https://github.com/facebookresearch/detectron2/blob/main/detectron2/config/defaults.py
                             for various settings.
    :param dataset_train: the dataset to use for training.
    :param path_weights: path to a checkpoint, if you want to continue training or fine-tune. Will train from scratch if
                         an empty string is passed
    :param config_overwrite: Pass a list of arguments if some configs from the .yaml file should be replaced. Use the
                             list convention, e.g. ['TRAIN.STEPS_PER_EPOCH=500', 'OUTPUT.RESULT_SCORE_THRESH=0.4']
    :param log_dir: Path to log dir. Will default to `train_log/frcnn`
    :param build_train_config: dataflow build setting. Again, use list convention setting, e.g. ['max_datapoints=1000']
    :param dataset_val: the dataset to use for validation.
    :param build_val_config: same as `build_train_config` but for validation
    :param metric_name: A metric name to choose for validation. Will use the default setting. If you want a custom
                        metric setting, pass a metric explicitly.
    :param metric: A metric to choose for validation.
    :param pipeline_component_name: A pipeline component name to use for validation.
    """

    assert get_num_gpu() > 0, "Has to train with GPU!"

    build_train_dict: Dict[str, str] = {}
    if build_train_config is not None:
        build_train_dict = string_to_dict(",".join(build_train_config))
    if "split" not in build_train_dict:
        build_train_dict["split"] = "train"

    build_val_dict: Dict[str, str] = {}
    if build_val_config is not None:
        build_val_dict = string_to_dict(",".join(build_val_config))
    if "split" not in build_val_dict:
        build_val_dict["split"] = "val"

    if config_overwrite is None:
        config_overwrite = []
    conf_list = [
        "MODEL.WEIGHTS",
        path_weights,
        "OUTPUT_DIR",
        log_dir,
    ]
    for conf in config_overwrite:
        key, val = conf.split("=", maxsplit=1)
        conf_list.extend([key, val])

    if isinstance(dataset_train, str):
        dataset_train = get_dataset(dataset_train)

    cfg = _set_config(
        path_config_yaml, conf_list, dataset_train, dataset_val, metric_name, metric, pipeline_component_name
    )

    if metric_name is not None:
        metric = metric_registry.get(metric_name)

    dataset = DatasetAdapter(dataset_train, True, image_to_d2_frcnn_training(False), True, **build_train_dict)
    augment_list = [ResizeShortestEdge(cfg.INPUT.MIN_SIZE_TRAIN, cfg.INPUT.MAX_SIZE_TRAIN), RandomFlip()]
    mapper = DatasetMapper(is_train=True, augmentations=augment_list, image_format="BGR")

    logger.info("Config: \n %s", str(cfg), cfg)

    trainer = D2Trainer(cfg, dataset, mapper)
    trainer.resume_or_load()

    if cfg.TEST.DO_EVAL:
        categories = dataset_val.dataflow.categories.get_categories(filtered=True)  # type: ignore
        config_overwrite = _update_for_eval(config_overwrite)
        detector = D2FrcnnDetector(path_config_yaml, path_weights, categories, config_overwrite, cfg.MODEL.DEVICE)
        pipeline_component_cls = pipeline_component_registry.get(pipeline_component_name)
        pipeline_component = pipeline_component_cls(detector)
        assert isinstance(pipeline_component, PredictorPipelineComponent)

        if metric_name is not None:
            metric = metric_registry.get(metric_name)
        assert metric is not None

        trainer.setup_evaluator(dataset_val, pipeline_component, metric, build_val_dict)  # type: ignore
    return trainer.train()



# -*- coding: utf-8 -*-
# File: test_segment.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module pipe.segment
"""

from typing import List

from pytest import mark

from deepdoctection.datapoint import BoundingBox, CategoryAnnotation, Image
from deepdoctection.extern.base import DetectionResult
from deepdoctection.pipe.segment import (
    PubtablesSegmentationService,
    SegmentationResult,
    TableSegmentationService,
    create_intersection_cells,
    stretch_items,
    tile_tables_with_items_per_table,
)
from deepdoctection.utils.settings import CellType, LayoutType


@mark.basic
def test_stretch_items(dp_image_tab_cell_item: Image, dp_image_item_stretched: Image) -> None:
    """test stretch_items"""
    # Arrange
    dp = dp_image_tab_cell_item
    dp_expected = dp_image_item_stretched
    table_name = LayoutType.table
    item_names = [LayoutType.row, LayoutType.column]

    # Act
    dp = stretch_items(dp, table_name, item_names[0], item_names[1], 0.001, 0.001)

    # Assert
    tables = dp.get_annotation(category_names=table_name)
    assert len(tables) == 1

    rows = dp.get_annotation(category_names=item_names[0])
    cols = dp.get_annotation(category_names=item_names[1])

    rows_expected = dp_expected.get_annotation(category_names=item_names[0])
    cols_expected = dp_expected.get_annotation(category_names=item_names[1])

    for row, row_expected in zip(rows, rows_expected):
        assert isinstance(row.image, Image)
        row_embedding = row.image.get_embedding(dp.image_id)
        assert isinstance(row_expected.image, Image)
        row_expected_embedding = row_expected.image.get_embedding(dp_expected.image_id)
        assert row_embedding == row_expected_embedding

    for col, col_expected in zip(cols, cols_expected):
        assert isinstance(col.image, Image)
        col_embedding = col.image.get_embedding(dp.image_id)
        assert isinstance(col_expected.image, Image)
        col_expected_embedding = col_expected.image.get_embedding(dp_expected.image_id)
        assert col_embedding == col_expected_embedding


class TestTableSegmentationService:
    """
    Test TableSegmentationService
    """

    def setup_method(self) -> None:
        """
        setup necessary components
        """

        self._segment_rule = "iou"
        self._iou_threshold_rows = 0.001
        self._iou_threshold_cols = 0.001
        self._ioa_threshold_rows = 0.4
        self._ioa_threshold_cols = 0.4
        self._remove_iou_threshold_rows = 0.001
        self._remove_iou_threshold_cols = 0.001
        self._tile_table_with_items = False

        self.table_segmentation_service = TableSegmentationService(
            self._segment_rule,  # type: ignore
            self._iou_threshold_rows if self._segment_rule in ["iou"] else self._ioa_threshold_rows,
            self._iou_threshold_cols if self._segment_rule in ["iou"] else self._ioa_threshold_cols,
            self._tile_table_with_items,
            self._remove_iou_threshold_rows,
            self._remove_iou_threshold_cols,
        )

    @mark.basic
    def test_pass_datapoint(self, dp_image_tab_cell_item: Image, dp_image_fully_segmented: Image) -> None:
        """test pass_datapoint"""

        # Arrange
        dp = dp_image_tab_cell_item
        dp_expected = dp_image_fully_segmented

        # Act
        dp = self.table_segmentation_service.pass_datapoint(dp)

        # Assert items have correctly assigned sub categories row/col number
        for item_name, sub_item_name in zip(
            self.table_segmentation_service._item_names,  # pylint: disable=W0212
            self.table_segmentation_service._sub_item_names,  # pylint: disable=W0212
        ):
            items = dp.get_annotation(category_names=item_name)
            items_expected = dp_expected.get_annotation(category_names=item_name)
            for el in zip(items, items_expected):
                item_cat = el[0].get_sub_category(sub_item_name)
                item_cat_expected = el[1].get_sub_category(sub_item_name)
                assert item_cat.category_name == item_cat_expected.category_name
                assert item_cat.category_id == item_cat_expected.category_id

        # Assert cells have correctly assigned sub categories row/col/rs/cs number
        cells = dp.get_annotation_iter(
            category_names=self.table_segmentation_service._cell_names  # pylint: disable=W0212
        )
        cells_expected = dp_expected.get_annotation_iter(
            category_names=self.table_segmentation_service._cell_names  # pylint: disable=W0212
        )

        for el in zip(cells, cells_expected):
            cell, cell_expected = el[0], el[1]
            row_sub_cat = cell.get_sub_category(CellType.row_number)
            row_sub_cat_expected = cell_expected.get_sub_category(CellType.row_number)
            assert row_sub_cat.category_name == row_sub_cat_expected.category_name
            assert row_sub_cat.category_id == row_sub_cat_expected.category_id

            col_sub_cat = cell.get_sub_category(CellType.column_number)
            col_sub_cat_expected = cell_expected.get_sub_category(CellType.column_number)
            assert col_sub_cat.category_name == col_sub_cat_expected.category_name
            assert col_sub_cat.category_id == col_sub_cat_expected.category_id

            rs_sub_cat = cell.get_sub_category(CellType.row_span)
            rs_sub_cat_expected = cell_expected.get_sub_category(CellType.row_span)
            assert rs_sub_cat.category_name == rs_sub_cat_expected.category_name
            assert rs_sub_cat.category_id == rs_sub_cat_expected.category_id

            cs_sub_cat = cell.get_sub_category(CellType.column_span)
            cs_sub_cat_expected = cell_expected.get_sub_category(CellType.column_span)
            assert cs_sub_cat.category_name == cs_sub_cat_expected.category_name
            assert cs_sub_cat.category_id == cs_sub_cat_expected.category_id


@mark.basic
def test_tile_tables_with_items_per_table(
    dp_image_item_stretched: Image,
    row_box_tiling_table: List[BoundingBox],
    col_box_tiling_table: List[BoundingBox],
    row_sub_cats: List[CategoryAnnotation],
    col_sub_cats: List[CategoryAnnotation],
) -> None:
    """test tile_tables_with_items_per_table"""

    # Arrange
    dp = dp_image_item_stretched
    rows = dp.get_annotation_iter(category_names=LayoutType.row)
    cols = dp.get_annotation_iter(category_names=LayoutType.column)

    for row, col, row_sub_cat, col_sub_cat in zip(rows, cols, row_sub_cats, col_sub_cats):
        row.dump_sub_category(CellType.row_number, row_sub_cat)
        col.dump_sub_category(CellType.column_number, col_sub_cat)

    table = dp.get_annotation(category_names=LayoutType.table)
    item_names = [LayoutType.row, LayoutType.column]  # row names must be before column name

    # Act
    dp = tile_tables_with_items_per_table(dp, table[0], item_names[0])
    dp = tile_tables_with_items_per_table(dp, table[0], item_names[1])

    # Assert
    rows = dp.get_annotation(category_names=LayoutType.row)
    cols = dp.get_annotation(category_names=LayoutType.column)

    assert rows[0].image is not None and rows[1].image is not None
    first_row_box = rows[0].image.get_embedding(dp.image_id)
    second_row_box = rows[1].image.get_embedding(dp.image_id)

    assert cols[0].image is not None and cols[1].image is not None
    first_col_box = cols[0].image.get_embedding(dp.image_id)
    second_col_box = cols[1].image.get_embedding(dp.image_id)

    assert first_row_box == row_box_tiling_table[0]
    assert second_row_box == row_box_tiling_table[1]

    assert first_col_box == col_box_tiling_table[0]
    assert second_col_box == col_box_tiling_table[1]


class TestTableSegmentationServiceWhenTableFullyTiled:
    """
    Test TableSegmentationService integrated when the table is fully tiled by rows and columns
    """

    def setup_method(self) -> None:
        """
        setup necessary components
        """

        self._segment_rule = "iou"
        self._iou_threshold_rows = 0.001
        self._iou_threshold_cols = 0.001
        self._ioa_threshold_rows = 0.4
        self._ioa_threshold_cols = 0.4
        self._remove_iou_threshold_rows = 0.001
        self._remove_iou_threshold_cols = 0.001
        self._tile_table_with_items = True

        self.tp_table_segmentation_service = TableSegmentationService(
            self._segment_rule,  # type: ignore
            self._iou_threshold_rows if self._segment_rule in ["iou"] else self._ioa_threshold_rows,
            self._iou_threshold_cols if self._segment_rule in ["iou"] else self._ioa_threshold_cols,
            self._tile_table_with_items,
            self._remove_iou_threshold_rows,
            self._remove_iou_threshold_cols,
        )

    @mark.basic
    def test_integration_pipeline_component(
        self, dp_image_tab_cell_item: Image, dp_image_fully_segmented_fully_tiled: Image
    ) -> None:
        """
        Integration test for pipeline component
        """

        # Arrange
        dp = dp_image_tab_cell_item
        dp_expected = dp_image_fully_segmented_fully_tiled

        # Act
        dp = self.tp_table_segmentation_service.pass_datapoint(dp)

        # Assert
        cells = dp.get_annotation(
            category_names=self.tp_table_segmentation_service._cell_names  # pylint: disable=W0212
        )

        cells_expected = dp_expected.get_annotation(
            category_names=self.tp_table_segmentation_service._cell_names  # pylint: disable=W0212
        )

        assert len(cells) == len(cells_expected)

        for cell, cell_expected in zip(cells, cells_expected):
            assert cell.get_sub_category(CellType.row_number) == cell_expected.get_sub_category(CellType.row_number)
            assert cell.get_sub_category(CellType.column_number) == cell_expected.get_sub_category(
                CellType.column_number
            )
            assert cell.get_sub_category(CellType.row_span) == cell_expected.get_sub_category(CellType.row_span)
            assert cell.get_sub_category(CellType.column_span) == cell_expected.get_sub_category(CellType.column_span)


@mark.basic
def test_create_intersection_cells(dp_image_tab_cell_item: Image) -> None:
    """
    Test create_intersection_cells generates cells from intersecting rows and columns and creates
    """

    # Arrange
    dp = dp_image_tab_cell_item

    rows = dp.get_annotation(category_names=LayoutType.row)
    cols = dp.get_annotation(category_names=LayoutType.column)
    for idx, items in enumerate(zip(rows, cols)):
        items[0].dump_sub_category(
            CellType.row_number, CategoryAnnotation(category_name=CellType.row_number, category_id=str(idx + 1))
        )
        items[1].dump_sub_category(
            CellType.column_number, CategoryAnnotation(category_name=CellType.column_number, category_id=str(idx + 1))
        )

    table = dp.get_annotation(category_names=LayoutType.table)[0]
    table_ann_id = table.annotation_id
    detect_result_cells, segment_result_cells = create_intersection_cells(
        rows, cols, table_ann_id, 5, [CellType.row_number, CellType.column_number]
    )
    expected_detect_result = [
        DetectionResult(box=[15.0, 100.0, 20.0, 150.0], class_id=5, class_name=LayoutType.cell),
        DetectionResult(box=[40.0, 100.0, 50.0, 150.0], class_id=5, class_name=LayoutType.cell),
        DetectionResult(box=[15.0, 200.0, 20.0, 240.0], class_id=5, class_name=LayoutType.cell),
        DetectionResult(box=[40.0, 200.0, 50.0, 240.0], class_id=5, class_name=LayoutType.cell),
    ]
    expected_segment_result = [
        SegmentationResult(row_num=1, col_num=1, rs=1, cs=1, annotation_id=""),
        SegmentationResult(row_num=1, col_num=2, rs=1, cs=1, annotation_id=""),
        SegmentationResult(row_num=2, col_num=1, rs=1, cs=1, annotation_id=""),
        SegmentationResult(row_num=2, col_num=2, rs=1, cs=1, annotation_id=""),
    ]

    assert len(detect_result_cells) == 4
    assert detect_result_cells == expected_detect_result
    assert len(segment_result_cells) == 4
    assert expected_segment_result == segment_result_cells


class TestPubtablesSegmentationService:
    """
    Test PubtablesSegmentationService
    """

    def setup_method(self) -> None:
        """
        setup necessary components
        """

        self._ioa_threshold_rows = 0.4
        self._ioa_threshold_cols = 0.4
        self._remove_iou_threshold_rows = 0.001
        self._remove_iou_threshold_cols = 0.001
        self._tile_table_with_items = True
        self.cell_class_id = 5

        self.table_segmentation_service = PubtablesSegmentationService(
            "ioa",
            self._ioa_threshold_rows,
            self._ioa_threshold_cols,
            self._tile_table_with_items,
            self._remove_iou_threshold_rows,
            self._remove_iou_threshold_cols,
            self.cell_class_id,
        )

    def test_pass_datapoint(self, dp_image_tab_cell_item: Image) -> None:
        """test pass_datapoint"""

        # Arrange
        dp = dp_image_tab_cell_item
        cells = dp.get_annotation(category_names=LayoutType.cell)
        table = dp.get_annotation(category_names=LayoutType.table)[0]

        for cell in cells:
            dp.remove(cell)

        tab_cells = table.image.get_annotation(category_names=LayoutType.cell)  # type: ignore
        for cell in tab_cells:
            table.image.remove(cell)  # type: ignore

        # Act
        dp = self.table_segmentation_service.pass_datapoint(dp)

        # Assert
        rows = dp.get_annotation(category_names=LayoutType.row)
        cols = dp.get_annotation(category_names=LayoutType.column)
        cells = dp.get_annotation(category_names=LayoutType.cell)

        assert len(rows) == 2
        assert len(cols) == 2
        assert len(cells) == 4



# -*- coding: utf-8 -*-
# File: pubtables1m.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for PubTables1M-Detection-PASCAL-VOC dataset. Install the dataset following the folder structure

    PubTables1M
    ├── PubTables1M-Detection-PASCAL-VOC
    │├── images
    ││ ├── PMC5700015_4.jpg
    ││ ├── PMC5700016_4.jpg
    │├── test
    ││ ├── PMC512281_8.xml
    ││ ├── PMC512281_9.xml
    │├── train
    ││ ├── ...
    │├── val
    ││ ├── ...
    │├── images_filelist.txt
    │├── test_filelist.txt
    │├── train_filelist.txt
    │├── val_filelist.txt
    ├── PubTables-1M-Structure_Annotations_Test
    ├── PubTables-1M-Structure_Images_Test
"""

import os
from typing import Mapping, Union

from ...dataflow import DataFlow, MapData, SerializerFiles
from ...datasets.info import DatasetInfo
from ...mapper.cats import filter_cat
from ...mapper.maputils import curry
from ...mapper.misc import xml_to_dict
from ...mapper.pascalstruct import pascal_voc_dict_to_image
from ...utils.detection_types import JsonDict
from ...utils.file_utils import lxml_available
from ...utils.settings import CellType, DatasetType, LayoutType
from ...utils.systools import get_package_path
from ..base import _BuiltInDataset
from ..dataflow_builder import DataFlowBaseBuilder
from ..info import DatasetCategories
from ..registry import dataset_registry

if lxml_available():
    from lxml import etree

_NAME = "pubtables1m_det"

_DESCRIPTION = (
    "[excerpt from Brandon Smock et. all. PubTables-1M: Towards Comprehensive Table Extraction From Unstructured \n"
    "Documents] '...we release PubTables1M, a dataset of nearly one million tables from PubMed Central Open Access \n"
    " scientific articles, with complete bounding box annotations for both table detection and structure \n"
    "recognition. In addition to being the largest dataset of its kind, PubTables1M addresses issues such as \n"
    " inherent ambiguity and lack of consistency in the source annotations, attempting to provide definitive ground \n"
    " truth labels through a thorough canonicalization and quality control process.' This dataset is devoted to two "
    "different tasks: table detection and table structure recognition. For this first task use  'pubtables1m_det' "
    "whereas for the second 'pubtables1m_struct'"
)

_LICENSE = "Community Data License Agreement – Permissive, Version 1.0"

_URL = "https://msropendata.com/datasets/505fcbe3-1383-42b1-913a-f651b8b712d3"

_SPLITS: Mapping[str, str] = {"train": "train", "val": "val", "test": "test"}
_TYPE = DatasetType.object_detection
_LOCATION = "PubTables1M"
_ANNOTATION_FILES: Mapping[str, str] = {
    "train": "PubTables1M-Detection-PASCAL-VOC/train",
    "val": "PubTables1M-Detection-PASCAL-VOC/val",
    "test": "PubTables1M-Detection-PASCAL-VOC/test",
}
_INIT_CATEGORIES_DET = [LayoutType.table, LayoutType.table_rotated]


@dataset_registry.register("pubtables1m_det")
class Pubtables1MDet(_BuiltInDataset):
    """
    Pubtables1MDet
    """

    _name = _NAME

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name=_NAME, description=_DESCRIPTION, license=_LICENSE, url=_URL, splits=_SPLITS, type=_TYPE)

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES_DET)

    def _builder(self) -> "Pubtables1MBuilder":
        return Pubtables1MBuilder(location=_LOCATION, annotation_files=_ANNOTATION_FILES)


class Pubtables1MBuilder(DataFlowBaseBuilder):
    """
    Pubtables1M dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the return
        values of the dataflow:

        `split:` Split of the dataset. Can be `train`, `val` or `test`. Default: `val`

        `max_datapoints:` Will stop iterating after max_datapoints. Default: `None`

        `load_image:` Will load the image for each datapoint.  Default: `False`

        `fake_score:` Will add a fake score so that annotations look like predictions

        :return: dataflow
        """

        split = str(kwargs.get("split", "val"))
        load_image = kwargs.get("load_image", False)
        max_datapoints = kwargs.get("max_datapoints")
        fake_score = kwargs.get("fake_score", False)

        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)

        # Load
        path_ann_files = self.get_workdir() / self.get_annotation_file(split)

        df = SerializerFiles.load(path_ann_files, ".xml", max_datapoints)
        utf8_parser = etree.XMLParser(encoding="utf-8")

        @curry
        def load_xml(path_ann: str, utf8_parser: etree.XMLParser) -> JsonDict:
            with open(path_ann, "r", encoding="utf-8") as xml_file:
                root = etree.fromstring(xml_file.read().encode("utf_8"), parser=utf8_parser)
            return {"file_name": path_ann, "xml": root}

        df = MapData(df, load_xml(utf8_parser))  # pylint: disable=E1120

        with open(
            os.path.join(get_package_path(), "deepdoctection/datasets/instances/xsl/pascal_voc.xsl"),
            "r",
            encoding="utf-8",
        ) as xsl_file:
            xslt_file = xsl_file.read().encode("utf-8")
        xml_obj = etree.XML(xslt_file, parser=etree.XMLParser(encoding="utf-8"))
        xslt_obj = etree.XSLT(xml_obj)

        df = MapData(df, xml_to_dict(xslt_obj))  # pylint: disable = E1120

        def _map_file_name(dp: JsonDict) -> JsonDict:
            path, file_name = os.path.split(dp["file_name"])
            path = os.path.join(os.path.split(path)[0], "images")
            dp["json"]["filename"] = os.path.join(path, file_name.replace(".xml", ".jpg"))
            return dp["json"]

        df = MapData(df, _map_file_name)
        df = MapData(
            df,
            pascal_voc_dict_to_image(  # pylint: disable = E1120
                self.categories.get_categories(init=True, name_as_key=True),
                load_image,
                filter_empty_image=True,
                fake_score=fake_score,
                category_name_mapping={"table": LayoutType.table, "table rotated": LayoutType.table_rotated},
            ),
        )

        return df


_NAME_STRUCT = "pubtables1m_struct"
_ANNOTATION_FILES_STRUCT: Mapping[str, str] = {
    "train": "PubTables-1M-Structure_Annotations_Train",
    "val": "PubTables-1M-Structure_Annotations_Val",
    "test": "PubTables-1M-Structure_Annotations_Test",
}

_INIT_CATEGORIES_STRUCT = [
    LayoutType.table,
    LayoutType.row,
    LayoutType.column,
    CellType.spanning,
    CellType.row_header,
    CellType.column_header,
    CellType.projected_row_header,
]

_IMAGES: Mapping[str, str] = {
    "train": "PubTables-1M-Structure_Images_Train",
    "val": "PubTables-1M-Structure_Images_Val",
    "test": "PubTables-1M-Structure_Images_Test",
}


@dataset_registry.register("pubtables1m_struct")
class Pubtables1MStruct(_BuiltInDataset):
    """
    Pubtables1MStruct
    """

    _name = _NAME_STRUCT

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(
            name=_NAME_STRUCT, description=_DESCRIPTION, license=_LICENSE, url=_URL, splits=_SPLITS, type=_TYPE
        )

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES_STRUCT)

    def _builder(self) -> "Pubtables1MBuilderStruct":
        return Pubtables1MBuilderStruct(location=_LOCATION, annotation_files=_ANNOTATION_FILES_STRUCT)


class Pubtables1MBuilderStruct(DataFlowBaseBuilder):
    """
    Pubtables1M dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the return
        values of the dataflow:

        `split:` Split of the dataset. Can be `train`, `val` or `test`. Default: `val`
        `max_datapoints:` Will stop iterating after max_datapoints. Default: `None`
        `load_image:` Will load the image for each datapoint.  Default: `False`
        `fake_score:` Will add a fake score so that annotations look like predictions

        :return: dataflow
        """

        split = str(kwargs.get("split", "val"))
        load_image = kwargs.get("load_image", False)
        max_datapoints = kwargs.get("max_datapoints")
        fake_score = kwargs.get("fake_score", False)

        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)

        # Load
        path_ann_files = self.get_workdir() / self.get_annotation_file(split)

        df = SerializerFiles.load(path_ann_files, ".xml", max_datapoints)
        utf8_parser = etree.XMLParser(encoding="utf-8")

        @curry
        def load_xml(path_ann: str, utf8_parser: etree.XMLParser) -> JsonDict:
            with open(path_ann, "r", encoding="utf-8") as xml_file:
                root = etree.fromstring(xml_file.read().encode("utf_8"), parser=utf8_parser)
            return {"file_name": path_ann, "xml": root}

        df = MapData(df, load_xml(utf8_parser))  # pylint: disable=E1120

        with open(
            os.path.join(get_package_path(), "deepdoctection/datasets/instances/xsl/pascal_voc.xsl"),
            "r",
            encoding="utf-8",
        ) as xsl_file:
            xslt_file = xsl_file.read().encode("utf-8")
        xml_obj = etree.XML(xslt_file, parser=etree.XMLParser(encoding="utf-8"))
        xslt_obj = etree.XSLT(xml_obj)

        df = MapData(df, xml_to_dict(xslt_obj))  # pylint: disable = E1120

        @curry
        def _map_file_name(dp: JsonDict, split: str) -> JsonDict:
            path, file_name = os.path.split(dp["file_name"])
            file_split = _IMAGES[split]
            path = os.path.join(os.path.split(path)[0], file_split)
            dp["json"]["filename"] = os.path.join(path, file_name.replace(".xml", ".jpg"))
            return dp["json"]

        df = MapData(df, _map_file_name(split))
        df = MapData(
            df,
            pascal_voc_dict_to_image(  # pylint: disable = E1120
                self.categories.get_categories(init=True, name_as_key=True),
                load_image,
                filter_empty_image=True,
                fake_score=fake_score,
                category_name_mapping={
                    "table": LayoutType.table,
                    "table spanning cell": CellType.spanning,
                    "table row": LayoutType.row,
                    "table row header": CellType.row_header,
                    "table projected row header": CellType.projected_row_header,
                    "table column": LayoutType.column,
                    "table column header": CellType.column_header,
                },
            ),
        )

        assert self.categories is not None  # avoid many typing issues
        if self.categories.is_filtered():
            df = MapData(
                df,
                filter_cat(
                    self.categories.get_categories(as_dict=False, filtered=True),
                    self.categories.get_categories(as_dict=False, filtered=False),
                ),
            )

        return df



# -*- coding: utf-8 -*-
# File: refine.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for refining methods of table segmentation. The refining methods lead ultimately to a table structure which
enables html table representations
"""
from collections import defaultdict
from copy import copy
from dataclasses import asdict
from itertools import chain, product
from typing import DefaultDict, List, Optional, Set, Tuple, Union

import networkx as nx  # type: ignore

from ..datapoint.annotation import ImageAnnotation
from ..datapoint.box import merge_boxes
from ..datapoint.image import Image
from ..extern.base import DetectionResult
from ..mapper.maputils import MappingContextManager
from ..utils.detection_types import JsonDict
from ..utils.settings import CellType, LayoutType, Relationships, TableType, get_type
from .base import PipelineComponent
from .registry import pipeline_component_registry

__all__ = ["TableSegmentationRefinementService", "generate_html_string"]


def tiles_to_cells(dp: Image, table: ImageAnnotation) -> List[Tuple[Tuple[int, int], str]]:
    """
    Creation of a table parquet: A table is divided into a tile parquet with the (number of rows) x
    (the number of columns) tiles.
    Each tile is assigned a list of cell ids that are occupied by the cell. No cells but one or more cells can be
    assigned per tile.

    :param dp: Image
    :param table: Table image annotation
    :return: Image
    """

    cell_ann_ids = table.get_relationship(Relationships.child)
    cells = dp.get_annotation(
        category_names=[LayoutType.cell, CellType.header, CellType.body], annotation_ids=cell_ann_ids
    )
    tile_to_cells = []

    for cell in cells:
        row_number = int(cell.get_sub_category(CellType.row_number).category_id)
        col_number = int(cell.get_sub_category(CellType.column_number).category_id)
        rs = int(cell.get_sub_category(CellType.row_span).category_id)
        cs = int(cell.get_sub_category(CellType.column_span).category_id)
        for k in range(rs):
            for l in range(cs):
                assert cell.annotation_id is not None, cell.annotation_id
                tile_to_cells.append(((row_number + k, col_number + l), cell.annotation_id))

    return tile_to_cells


def connected_component_tiles(
    tile_to_cell_list: List[Tuple[Tuple[int, int], str]]
) -> Tuple[List[Set[Tuple[int, int]]], DefaultDict[Tuple[int, int], List[str]]]:
    """
    The assignment of bricks to their cell occupancy induces a graph, with bricks as corners and cell edges. Cells that
    lie on top of several bricks connect the underlying bricks. The graph generated according to this procedure is
    usually multiple connected. The related components and the tile/cell ids assignment are determined.

    :param tile_to_cell_list: List of tuples with tile position and cell ids
    :return: List of set with tiles that belong to the same connected component and a dict with tiles as keys and
             assigned list of cell ids as values.
    """
    cell_to_tile_list = [(cell_position[1], cell_position[0]) for cell_position in tile_to_cell_list]
    cells = set(tup[0] for tup in cell_to_tile_list)

    tile_to_cell_dict = defaultdict(list)
    for key, val in tile_to_cell_list:
        tile_to_cell_dict[key].append(val)

    cell_to_tile_dict = defaultdict(list)
    for key, val in cell_to_tile_list:  # type: ignore
        cell_to_tile_dict[key].append(val)

    cell_to_edges = defaultdict(list)
    for key, _ in tile_to_cell_dict.items():
        cell_to_edges[key].extend(
            list(set(product(tile_to_cell_dict[key], tile_to_cell_dict[key])))  # pylint: disable=R1733
        )
    graph = nx.Graph()
    graph.add_nodes_from(cells)
    graph.add_edges_from(chain(*cell_to_edges.values()))
    connected_components_cell = list(nx.connected_components(graph))
    connected_components_tiles = []

    for component in connected_components_cell:
        tiles: Set[Tuple[int, int]] = set()
        for cell in component:
            tiles = tiles.union(set(cell_to_tile_dict[cell]))  # type: ignore
        connected_components_tiles.append(tiles)

    return connected_components_tiles, tile_to_cell_dict


def _missing_tile(inputs: Set[Tuple[int, int]]) -> Optional[Tuple[int, int]]:
    min_x, min_y, max_x, max_y = (
        min(a[0] for a in inputs),
        min(a[1] for a in inputs),
        max(a[0] for a in inputs),
        max(a[1] for a in inputs),
    )

    for x in range(min_x, max_x):
        for y in range(min_y, max_y):
            if (x, y) not in inputs:
                return (x, y)
    return None


def _find_component(
    tile: Tuple[int, int], reduced_connected_tiles: List[Set[Tuple[int, int]]]
) -> Optional[Set[Tuple[int, int]]]:
    for comp in reduced_connected_tiles:
        if tile in comp:
            return comp
    return None


def _merge_components(reduced_connected_tiles: List[Set[Tuple[int, int]]]) -> List[Set[Tuple[int, int]]]:
    new_reduced_connected_tiles = []
    for connected_tile in reduced_connected_tiles:
        out = _missing_tile(connected_tile)
        if out is not None:
            component = _find_component(out, reduced_connected_tiles)
            if component is not None:
                new_connected_tile = connected_tile.union(component)
                reduced_connected_tiles.remove(component)
                reduced_connected_tiles.remove(connected_tile)
                new_reduced_connected_tiles.append(new_connected_tile)
                if component in new_reduced_connected_tiles:
                    new_reduced_connected_tiles.remove(component)
            else:
                new_connected_tile = connected_tile.union({out})
                new_reduced_connected_tiles.append(new_connected_tile)
        else:
            new_reduced_connected_tiles.append(connected_tile)

    return new_reduced_connected_tiles


def generate_rectangle_tiling(connected_components_tiles: List[Set[Tuple[int, int]]]) -> List[Set[Tuple[int, int]]]:
    """
    The determined connected components imply that all cells have to be combined which are above a connected component.
    In addition, however, it must also be taken into account that cells must be rectangular. This means that related
    components have to be combined whose combined cells above do not create a rectangular tiling. All tiles are combined
    in such a way that all cells above them combine to form a rectangular scheme.

    :param connected_components_tiles: List of set with tiles that belong to the same connected component
    :return: List of sets with tiles, the cells on top of which together form a rectangular scheme
    """
    rectangle_tiling: List[Set[Tuple[int, int]]] = []
    inputs = connected_components_tiles

    while rectangle_tiling != inputs:
        if rectangle_tiling:
            inputs = rectangle_tiling
        rectangle_tiling = _merge_components(copy(inputs))

    return rectangle_tiling


def rectangle_cells(
    rectangle_tiling: List[Set[Tuple[int, int]]], tile_to_cell_dict: DefaultDict[Tuple[int, int], List[str]]
) -> List[Set[str]]:
    """
    All cells are determined that are located above combined connected components and form a rectangular scheme.

    :param rectangle_tiling: List of sets with tiles, the cells on top of which together form a rectangular scheme
    :param tile_to_cell_dict: Dict with tiles as keys and assigned list of cell ids as values.
    :return: List of set of cell ids that form a rectangular scheme
    """
    rectangle_tiling_cells: List[Set[str]] = []
    for rect_tiling_component in rectangle_tiling:
        rect_cell_component: Set[str] = set()
        for el in rect_tiling_component:
            rect_cell_component = rect_cell_component.union(set(tile_to_cell_dict[el]))
        rectangle_tiling_cells.append(rect_cell_component)
    return rectangle_tiling_cells


def _tiling_to_cell_position(inputs: Set[Tuple[int, int]]) -> Tuple[int, int, int, int]:
    row_number = min(a[0] for a in inputs)
    col_number = min(a[1] for a in inputs)
    row_span = max(abs(a[0] - b[0]) + 1 for a in inputs for b in inputs)
    col_span = max(abs(a[1] - b[1]) + 1 for a in inputs for b in inputs)
    return int(row_number), int(col_number), int(row_span), int(col_span)


def _html_cell(
    cell_position: Union[Tuple[int, int, int, int], Tuple[()]], position_filled_list: List[Tuple[int, int]]
) -> List[str]:
    """
    Html table cell string generation
    """
    html = ["<td"]
    if not cell_position:
        pass
    else:
        if cell_position[2] != 1:
            html.append(f" rowspan={cell_position[2]}")
        if cell_position[3] != 1:
            html.append(f" colspan={cell_position[3]}")
        if cell_position[2] != 1 or cell_position[3] != 1:
            position_filled_list.extend(
                [
                    (cell_position[0] + h_shift, cell_position[1] + v_shift)
                    for h_shift in range(cell_position[2])
                    for v_shift in range(cell_position[3])
                ]
            )
    html.append(">")
    str_html = "".join(html)
    html_list = [str_html, "</td>"]
    return html_list


def _html_row(
    row_list: List[Tuple[int, int, int, int]],
    position_filled_list: List[Tuple[int, int]],
    this_row: int,
    number_of_cols: int,
    row_ann_id_list: List[str],
) -> List[str]:
    """
    Html table row string generation
    """
    html = ["<tr>"]
    for idx in range(1, number_of_cols + 1):
        position_filled_this_row = list(filter(lambda x: x[0] == this_row, position_filled_list))
        column_filled_this_row = list(zip(*position_filled_this_row))
        column_filled_this_row = (
            [column_filled_this_row, column_filled_this_row]  # type:ignore
            if not column_filled_this_row
            else column_filled_this_row
        )
        if idx in column_filled_this_row[1]:
            pass
        else:
            cell_position_list = list(filter(lambda x: x[1] == idx, row_list))  # pylint:disable=W0640

            if cell_position_list:
                cell_position = cell_position_list[0]
                cell_id = row_ann_id_list.pop(0)
                ret_html = _html_cell(cell_position, position_filled_list)
                ret_html.insert(1, cell_id)
            else:
                cell_position = ()  # type: ignore
                ret_html = _html_cell(cell_position, position_filled_list)
            html.extend(ret_html)
    html.append("</tr>")
    return html


def _html_table(
    table_list: List[Tuple[int, List[Tuple[int, int, int, int]]]],
    cells_ann_list: List[Tuple[int, List[str]]],
    number_of_rows: int,
    number_of_cols: int,
) -> List[str]:
    """
    Html table string generation
    """
    html = ["<table>"]
    position_filled: List[Tuple[int, int]] = []
    for idx in range(1, number_of_rows + 1):
        row_idx = list(filter(lambda x: x[0] == idx, table_list))[0][1]  # pylint:disable=W0640
        row_ann_ids = list(filter(lambda x: x[0] == idx, cells_ann_list))[0][1]  # pylint:disable=W0640
        ret_html = _html_row(row_idx, position_filled, idx, number_of_cols, row_ann_ids)
        html.extend(ret_html)
    html.append("</table>")
    return html


def generate_html_string(table: ImageAnnotation) -> List[str]:
    """
    Takes the table segmentation by using table cells row number, column numbers etc. and generates a html
    representation.

    :param table: An annotation that has a not None image and fully segmented cell annotation.
    :return: HTML representation of the table
    """
    if table.image is None:
        raise ValueError("table.image cannot be None")
    table_image = table.image
    cells = table_image.get_annotation(
        category_names=[
            LayoutType.cell,
            CellType.header,
            CellType.body,
            CellType.spanning,
            CellType.row_header,
            CellType.column_header,
            CellType.projected_row_header,
        ]
    )
    number_of_rows = int(table_image.summary.get_sub_category(TableType.number_of_rows).category_id)
    number_of_cols = int(table_image.summary.get_sub_category(TableType.number_of_columns).category_id)
    table_list = []
    cells_ann_list = []
    for row_number in range(1, number_of_rows + 1):
        cells_of_row = list(
            sorted(
                filter(
                    lambda cell: cell.get_sub_category(CellType.row_number).category_id
                    == str(row_number),  # pylint: disable=W0640
                    cells,
                ),
                key=lambda cell: cell.get_sub_category(CellType.column_number).category_id,
            )
        )
        row_list = [
            (
                int(cell.get_sub_category(CellType.row_number).category_id),
                int(cell.get_sub_category(CellType.column_number).category_id),
                int(cell.get_sub_category(CellType.row_span).category_id),
                int(cell.get_sub_category(CellType.column_span).category_id),
            )
            for cell in cells_of_row
        ]
        ann_list = [cell.annotation_id for cell in cells_of_row]
        table_list.append((row_number, row_list))
        cells_ann_list.append((row_number, ann_list))
    return _html_table(table_list, cells_ann_list, number_of_rows, number_of_cols)


@pipeline_component_registry.register("TableSegmentationRefinementService")
class TableSegmentationRefinementService(PipelineComponent):
    """
    Refinement of the cell segmentation. The aim of this component is to create a table structure so that an HTML
    structure can be created.

    Assume that the arrangement of cells, rows and in the table is as follows in the original state. There is only one
    column.

    +----------+
    | C1   C2  |
    +----------+
    | C3   C3  |
    +----------+

    The first two cells have the same column assignment via the segmentation and must therefore be merged.
    Note that the number of rows and columns does not change in the refinement process. What changes is just the number
    of cells.

    Furthermore, when merging, it must be ensured that the combined cells still have a rectangular shape.
    This is also guaranteed in the refining process.

    +----------+
    | C1 |  C2 |
    +          +
    | C3 |  C3 |
    +----------+

    The table consists of one row and two columns. The upper cells belong together with the lower cell.
    However, this means that all cells must be merged with one another so that the table only consists of one cell
    after the refinement process.

    **Example**

            layout = ImageLayoutService(layout_detector, to_image=True, crop_image=True)
            cell = SubImageLayoutService(cell_detector, "TABLE")
            row_col = SubImageLayoutService(row_col_detector, "TABLE")
            table_segmentation = TableSegmentationService("ioa",0.9,0.8,True,0.0001,0.0001)
            table_segmentation_refinement = TableSegmentationRefinementService()

            table_recognition_pipe = DoctectionPipe([layout,
                                                     cell,
                                                     row_col,
                                                     table_segmentation,
                                                     table_segmentation_refinement])
            df = pipe.analyze(path="path/to/document.pdf")

            for dp in df:
                ...

    """

    def __init__(self) -> None:
        self._table_name = LayoutType.table
        self._cell_names = [CellType.header, CellType.body, LayoutType.cell]
        super().__init__("table_segment_refine")

    def serve(self, dp: Image) -> None:
        tables = dp.get_annotation(category_names=self._table_name)
        for table in tables:
            if table.image is None:
                raise ValueError("table.image cannot be None")
            tiles_to_cells_list = tiles_to_cells(dp, table)
            connected_components, tile_to_cell_dict = connected_component_tiles(tiles_to_cells_list)
            rectangle_tiling = generate_rectangle_tiling(connected_components)
            rectangle_cells_list = rectangle_cells(rectangle_tiling, tile_to_cell_dict)
            for tiling, cells_to_merge in zip(rectangle_tiling, rectangle_cells_list):
                no_context_error = True
                if len(cells_to_merge) != 1:
                    cells = dp.get_annotation(annotation_ids=list(cells_to_merge))
                    cell_boxes = [
                        cell.image.get_embedding(table.image.image_id) for cell in cells if cell.image is not None
                    ]
                    merged_box = merge_boxes(*cell_boxes)
                    det_result = DetectionResult(
                        box=merged_box.to_list(mode="xyxy"),
                        score=-1.0,
                        class_id=int(cells[0].category_id),
                        class_name=get_type(cells[0].category_name),
                    )
                    new_cell_ann_id = self.dp_manager.set_image_annotation(det_result, table.annotation_id)
                    if new_cell_ann_id is not None:
                        row_number, col_number, row_span, col_span = _tiling_to_cell_position(tiling)
                        self.dp_manager.set_category_annotation(
                            CellType.row_number, row_number, CellType.row_number, new_cell_ann_id
                        )
                        self.dp_manager.set_category_annotation(
                            CellType.column_number, col_number, CellType.column_number, new_cell_ann_id
                        )
                        self.dp_manager.set_category_annotation(
                            CellType.row_span, row_span, CellType.row_span, new_cell_ann_id
                        )
                        self.dp_manager.set_category_annotation(
                            CellType.column_span, col_span, CellType.column_span, new_cell_ann_id
                        )
                    else:
                        # DetectionResult cannot be dumped, hence merged_box must already exist. Hence, it must
                        # contain all other boxes. Hence, we must deactivate all other boxes.
                        with MappingContextManager(
                            dp_name=dp.file_name, filter_level="annotation", detect_result=asdict(det_result)
                        ) as annotation_context:
                            box_index = cell_boxes.index(merged_box)
                            cells.pop(box_index)
                        no_context_error = not annotation_context.context_error
                    if no_context_error:
                        for cell in cells:
                            cell.deactivate()

            cells = table.image.get_annotation(category_names=self._cell_names)
            number_of_rows = max([int(cell.get_sub_category(CellType.row_number).category_id) for cell in cells])
            number_of_cols = max([int(cell.get_sub_category(CellType.column_number).category_id) for cell in cells])
            max_row_span = max([int(cell.get_sub_category(CellType.row_span).category_id) for cell in cells])
            max_col_span = max([int(cell.get_sub_category(CellType.column_span).category_id) for cell in cells])
            # TODO: the summaries should be sub categories of the underlying ann
            self.dp_manager.set_summary_annotation(
                TableType.number_of_rows, TableType.number_of_rows, number_of_rows, annotation_id=table.annotation_id
            )
            self.dp_manager.set_summary_annotation(
                TableType.number_of_columns,
                TableType.number_of_columns,
                number_of_cols,
                annotation_id=table.annotation_id,
            )
            self.dp_manager.set_summary_annotation(
                TableType.max_row_span, TableType.max_row_span, max_row_span, annotation_id=table.annotation_id
            )
            self.dp_manager.set_summary_annotation(
                TableType.max_col_span, TableType.max_col_span, max_col_span, annotation_id=table.annotation_id
            )
            html = generate_html_string(table)
            self.dp_manager.set_container_annotation(TableType.html, -1, TableType.html, table.annotation_id, html)

    def clone(self) -> PipelineComponent:
        return self.__class__()

    def get_meta_annotation(self) -> JsonDict:
        return dict(
            [
                ("image_annotations", []),
                (
                    "sub_categories",
                    {
                        LayoutType.cell: {
                            CellType.row_number,
                            CellType.column_number,
                            CellType.row_span,
                            CellType.column_span,
                        },
                        LayoutType.table: {TableType.html},
                    },
                ),
                ("relationships", {}),
                ("summaries", []),
            ]
        )



# -*- coding: utf-8 -*-
# File: test_tedsmetric.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module eval.tedsmetric
"""

from pytest import mark

from deepdoctection.utils.file_utils import apted_available

if apted_available():
    from deepdoctection.eval.tedsmetric import teds_metric


@mark.basic
def test_teds_metric_returns_correct_distance() -> None:
    """
    teds returns score of 1.0 when comparing identical html strings
    """

    html_str = "<table><tr><td></td><td></td><td></td></tr><tr><td></td><td></td><td></td></tr></table>"
    results, number_results = teds_metric([html_str], [html_str], False)

    assert number_results == 1
    assert results == 1.0



# -*- coding: utf-8 -*-
# File: test_tpdetect.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module extern.tpdetect
"""

from typing import Dict, List
from unittest.mock import MagicMock, patch

from pytest import mark, raises

from deepdoctection.extern.base import DetectionResult
from deepdoctection.utils.detection_types import ImageType
from deepdoctection.utils.file_utils import tensorpack_available, tf_available
from deepdoctection.utils.settings import ObjectTypes

if tf_available() and tensorpack_available():
    from deepdoctection.extern.tp.tpfrcnn.modeling.generalized_rcnn import ResNetFPNModel
    from deepdoctection.extern.tpdetect import TPFrcnnDetector


def set_num_gpu_to_zero() -> int:
    """
    set gpu number to zero
    """
    return 0


def set_num_gpu_to_one() -> int:
    """
    set gpu number to one
    """
    return 1


def get_mock_detection_results(  # type: ignore
    np_img: ImageType,  # pylint: disable=W0613
    predictor,  # pylint: disable=W0613
    preproc_short_edge_size,  # pylint: disable=W0613
    preproc_max_size,  # pylint: disable=W0613
    mrcnn_accurate_paste,  # pylint: disable=W0613
) -> List[DetectionResult]:
    """
    returns list with two DetectionResult
    """

    return [
        DetectionResult(box=[1.0, 1.6, 2.0, 4.6], score=0.97, class_id=2),
        DetectionResult(box=[12.0, 12.0, 12.0, 12.0], score=0.03, class_id=4),
    ]


class TestTPFrcnnDetector:
    """
    Test TPFrcnnDetector constructor
    """

    @staticmethod
    @mark.requires_tf
    @patch("deepdoctection.extern.tp.tpcompat.get_num_gpu", MagicMock(side_effect=set_num_gpu_to_zero))
    def test_tp_frcnn_does_not_build_when_no_gpu(
        path_to_tp_frcnn_yaml: str, categories: Dict[str, ObjectTypes]
    ) -> None:
        """
        TP FRCNN needs one GPU for predicting. Construction fails, when no GPU is found
        """
        # Arrange, Act & Assert
        with raises(AssertionError):
            TPFrcnnDetector(path_yaml=path_to_tp_frcnn_yaml, path_weights="", categories=categories)

    @staticmethod
    @mark.requires_tf
    @patch("deepdoctection.extern.tp.tpcompat.get_num_gpu", MagicMock(side_effect=set_num_gpu_to_one))
    def test_tp_frcnn_returns_fpn_model(path_to_tp_frcnn_yaml: str, categories: Dict[str, ObjectTypes]) -> None:
        """
        TP FRCNN builds RestNetFPN model is construction is successful.
        """
        # Arrange, Act
        frcnn = TPFrcnnDetector(path_yaml=path_to_tp_frcnn_yaml, path_weights="", categories=categories)

        # Assert
        assert isinstance(frcnn._model, ResNetFPNModel)  # pylint: disable=W0212

    @staticmethod
    @mark.requires_tf
    @patch("deepdoctection.extern.tp.tpcompat.get_num_gpu", MagicMock(side_effect=set_num_gpu_to_one))
    @patch("deepdoctection.extern.tp.tpcompat.TensorpackPredictor._build_config", MagicMock())
    @patch("deepdoctection.extern.tp.tpcompat.TensorpackPredictor.get_predictor", MagicMock())
    @patch("deepdoctection.extern.tpdetect.tp_predict_image", MagicMock(side_effect=get_mock_detection_results))
    def test_tp_frcnn_predicts_image(
        path_to_tp_frcnn_yaml: str, categories: Dict[str, ObjectTypes], np_image: ImageType
    ) -> None:
        """
        TP FRCNN calls predict_image and post processes DetectionResult correctly, e.g. adding class names
        """

        # Arrange
        frcnn = TPFrcnnDetector(path_yaml=path_to_tp_frcnn_yaml, path_weights="", categories=categories)

        # Act
        results = frcnn.predict(np_image)

        # Assert
        assert len(results) == 2
        first_detect_result = results[0]
        second_detect_result = results[1]
        assert first_detect_result.class_name == categories["2"]
        assert second_detect_result.class_name == categories["4"]



# -*- coding: utf-8 -*-
# File: doclaynet.py

# Copyright 2022 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for DocLayNet dataset. Place the dataset as follows

    DocLayNet_core
    ├── COCO
    │ ├── test.json
    │ ├── val.json
    ├── PNG
    │ ├── 0a0d43e301facee9e99cc33b9b16e732dd207135f4027e75f6aea2bf117535a2.png
"""

import os
from typing import Mapping, Sequence, Union

from ...dataflow import DataFlow, MapData, MapDataComponent, SerializerCoco
from ...datapoint.annotation import CategoryAnnotation, SummaryAnnotation
from ...datapoint.image import Image
from ...mapper.cats import add_summary, cat_to_sub_cat, filter_cat, filter_summary
from ...mapper.cocostruct import coco_to_image
from ...mapper.maputils import curry
from ...utils.detection_types import JsonDict
from ...utils.fs import load_image_from_file
from ...utils.settings import DatasetType, DocumentType, LayoutType, ObjectTypes, PageType, TypeOrStr
from ..base import DatasetBase
from ..dataflow_builder import DataFlowBaseBuilder
from ..info import DatasetCategories, DatasetInfo
from ..registry import dataset_registry

_NAME = "doclaynet"
_DESCRIPTION = (
    "DocLayNet is a human-annotated document layout segmentation dataset containing 80863 pages from a broad "
    "variety of document sources. \n"
    "DocLayNet provides page-by-page layout segmentation ground-truth using bounding-boxes for 11 distinct class"
    " labels on 80863 unique pages from 6 document categories. It provides several unique features compared"
    " to related work such as PubLayNet or DocBank: \n"
    "Human Annotation: DocLayNet is hand-annotated by well-trained experts, providing a gold-standard in layout"
    " segmentation through human recognition and interpretation of each page layout \n"
    "Large layout variability: DocLayNet includes diverse and complex layouts from a large variety of public"
    " sources in Finance, Science, Patents, Tenders, Law texts and Manuals \n"
    "Detailed label set: DocLayNet defines 11 class labels to distinguish layout features in high detail. \n"
    "Redundant annotations: A fraction of the pages in DocLayNet are double- or triple-annotated, allowing to"
    " estimate annotation uncertainty and an upper-bound of achievable prediction accuracy with ML models "
    "Pre-defined train- test- and validation-sets: DocLayNet provides fixed sets for each to ensure proportional"
    " representation of the class-labels and avoid leakage of unique layout styles across the sets."
)
_LICENSE = "CDLA-Permissive"
_URL = "https://codait-cos-dax.s3.us.cloud-object-storage.appdomain.cloud/dax-doclaynet/1.0.0/DocLayNet_core.zip"
_SPLITS: Mapping[str, str] = {"train": "train", "val": "val", "test": "test"}
_TYPE = DatasetType.object_detection

_LOCATION = "DocLayNet_core"

_ANNOTATION_FILES: Mapping[str, str] = {"train": "COCO/train.json", "val": "COCO/val.json", "test": "COCO/test.json"}
_INIT_CATEGORIES = [
    LayoutType.caption,
    LayoutType.footnote,
    LayoutType.formula,
    LayoutType.list,
    LayoutType.page_footer,
    LayoutType.page_header,
    LayoutType.figure,
    LayoutType.section_header,
    LayoutType.table,
    LayoutType.text,
    LayoutType.title,
]
_SUB_CATEGORIES: Mapping[ObjectTypes, Mapping[ObjectTypes, Sequence[ObjectTypes]]] = {
    LayoutType.caption: {DatasetType.publaynet: [LayoutType.text]},
    LayoutType.footnote: {DatasetType.publaynet: [LayoutType.text]},
    LayoutType.formula: {DatasetType.publaynet: [LayoutType.text]},
    LayoutType.list: {DatasetType.publaynet: [LayoutType.list]},
    LayoutType.page_footer: {DatasetType.publaynet: [LayoutType.text]},
    LayoutType.page_header: {DatasetType.publaynet: [LayoutType.title]},
    LayoutType.figure: {DatasetType.publaynet: [LayoutType.figure]},
    LayoutType.section_header: {DatasetType.publaynet: [LayoutType.title]},
    LayoutType.table: {DatasetType.publaynet: [LayoutType.table]},
    LayoutType.text: {DatasetType.publaynet: [LayoutType.text]},
    LayoutType.title: {DatasetType.publaynet: [LayoutType.title]},
}


@dataset_registry.register("doclaynet")
class DocLayNet(DatasetBase):
    """
    DocLayNetSeq
    """

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name=_NAME, description=_DESCRIPTION, license=_LICENSE, splits=_SPLITS, type=_TYPE, url=_URL)

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES, init_sub_categories=_SUB_CATEGORIES)

    def _builder(self) -> "DocLayNetBuilder":
        return DocLayNetBuilder(location=_LOCATION, annotation_files=_ANNOTATION_FILES)


class DocLayNetBuilder(DataFlowBaseBuilder):
    """
    DocLayNetBuilder dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the returns
        of the dataflow:

        `split:` Split of the dataset. Can be `train`,`val` or `test`. Default: `val`

        `max_datapoints:` Will stop iterating after max_datapoints. Default: None

        `load_image:` Will load the image for each datapoint.  Default: False

        `fake_score:` Will add a fake score so that annotations look like predictions

        :return: dataflow
        """
        split = str(kwargs.get("split", "val"))
        max_datapoints = kwargs.get("max_datapoints")
        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)
        load_image = kwargs.get("load_image", False)
        fake_score = kwargs.get("fake_score", False)
        resized = kwargs.get("resized", False)

        # Load
        df: DataFlow
        path = self.get_workdir() / self.get_annotation_file(split)
        df = SerializerCoco.load(path, max_datapoints=max_datapoints)

        # Map
        if resized:
            png_folder = "PNG_resized"
        else:
            png_folder = "PNG"

        df = MapDataComponent(df, lambda dp: self.get_workdir() / png_folder / dp, "file_name")
        df = MapData(
            df,
            coco_to_image(
                self.categories.get_categories(init=True),
                load_image,
                filter_empty_image=True,
                fake_score=fake_score,
                coarse_mapping={1: 10, 2: 10, 3: 10, 4: 4, 5: 10, 6: 11, 7: 7, 8: 11, 9: 9, 10: 10, 11: 11},
                coarse_sub_cat_name=DatasetType.publaynet,
            ),
        )

        if self.categories.is_cat_to_sub_cat():
            df = MapData(
                df,
                cat_to_sub_cat(self.categories.get_categories(name_as_key=True), self.categories.cat_to_sub_cat),
            )

        assert self.categories is not None  # avoid many typing issues
        if self.categories.is_filtered():
            df = MapData(
                df,
                filter_cat(
                    self.categories.get_categories(as_dict=False, filtered=True),
                    self.categories.get_categories(as_dict=False, filtered=False),
                ),
            )
        df = MapData(df, add_summary(self.categories.get_categories(filtered=True)))
        return df


_NAME_SEQ = "doclaynet-seq"
_TYPE_SEQ = DatasetType.sequence_classification
_INIT_CATEGORIES_SEQ = [
    DocumentType.financial_report,
    DocumentType.scientific_publication,
    DocumentType.laws_and_regulations,
    DocumentType.government_tenders,
    DocumentType.manuals,
    DocumentType.patents,
]


@dataset_registry.register("doclaynet-seq")
class DocLayNetSeq(DatasetBase):
    """
    DocLayNetSeq is the DocLayNet dataset where the dataflow has been prepared to perform sequence classification
    """

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name=_NAME_SEQ, description=_DESCRIPTION, license=_LICENSE, splits=_SPLITS, type=_TYPE_SEQ)

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES_SEQ)

    def _builder(self) -> "DocLayNetSeqBuilder":
        return DocLayNetSeqBuilder(location=_LOCATION, annotation_files=_ANNOTATION_FILES)


class DocLayNetSeqBuilder(DataFlowBaseBuilder):
    """
    DocLayNetSeqBuilder dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the returns
        of the dataflow:

        `split:` Split of the dataset. Can be `train`, `val` or `test`. Default: `val`
        `max_datapoints:` Will stop iterating after max_datapoints. Default: `None`
        `load_image:` Will load the image for each datapoint.  Default: `False`

        :return: dataflow
        """
        split = str(kwargs.get("split", "val"))
        max_datapoints = kwargs.get("max_datapoints")
        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)
        load_image = kwargs.get("load_image", False)

        # Load
        dataset_split = self.annotation_files[split]
        path = self.get_workdir() / dataset_split  # type: ignore
        df = SerializerCoco.load(path, max_datapoints=max_datapoints)

        # Map
        df = MapDataComponent(df, lambda dp: self.get_workdir() / "PNG" / dp, "file_name")

        @curry
        def _map_to_image(dp: JsonDict, load_img: bool) -> Image:
            image = Image(location=dp["file_name"], file_name=os.path.split(dp["file_name"])[1])
            image.image = load_image_from_file(image.location)
            summary = SummaryAnnotation()
            label_to_category_name = {
                "financial_reports": DocumentType.financial_report,
                "scientific_articles": DocumentType.scientific_publication,
                "laws_and_regulations": DocumentType.laws_and_regulations,
                "government_tenders": DocumentType.government_tenders,
                "manuals": DocumentType.manuals,
                "patents": DocumentType.patents,
            }
            categories_dict = self.categories.get_categories(init=True, name_as_key=True)
            category_name = label_to_category_name[dp["doc_category"]]
            summary.dump_sub_category(
                PageType.document_type,
                CategoryAnnotation(category_name=category_name, category_id=categories_dict[category_name]),
            )
            image.summary = summary
            if not load_img:
                image.clear_image()
            return image

        df = MapData(df, _map_to_image(load_image))

        assert self.categories is not None
        if self.categories.is_filtered():
            df = MapData(
                df,
                filter_summary({PageType.document_type: self.categories.get_categories(as_dict=False, filtered=True)}),
            )

            @curry
            def _re_map_cat_ids(dp: Image, filtered_categories_name_as_key: Mapping[TypeOrStr, str]) -> Image:
                if dp.summary:
                    if PageType.document_type in dp.summary.sub_categories:
                        summary_cat = dp.summary.get_sub_category(PageType.document_type)
                        summary_cat.category_id = filtered_categories_name_as_key[summary_cat.category_name]
                return dp

            df = MapData(df, _re_map_cat_ids(self.categories.get_categories(filtered=True, name_as_key=True)))

        return df



# -*- coding: utf-8 -*-
# File: funsd.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for Funsd dataset.  Install the dataset following the folder structure

    funsd
    ├── testing_data
    │ ├── annotations
    │ │ ├── 82092117.json
    │ │ ├── 82092117_0338.json
    │ ├── images
    │ │ ├── 82092117.png
    │ │ ├── 82092117_0338.png
    ├── training_data
    │ ├── annotations
    │ │ ├── ...
    │ ├── images
    │ │ ├── ...
"""

import os
from typing import Dict, List, Mapping, Union

from ...dataflow import DataFlow, MapData, SerializerFiles
from ...datasets.info import DatasetInfo
from ...mapper.cats import cat_to_sub_cat, filter_cat
from ...mapper.xfundstruct import xfund_to_image
from ...utils.detection_types import JsonDict, Pathlike
from ...utils.fs import load_json
from ...utils.settings import BioTag, DatasetType, LayoutType, ObjectTypes, TokenClasses, TokenClassWithTag, WordType
from ..base import _BuiltInDataset
from ..dataflow_builder import DataFlowBaseBuilder
from ..info import DatasetCategories
from ..registry import dataset_registry


def load_file(path_ann: Pathlike) -> JsonDict:
    """
    Loading json file

    :param path_ann: path
    :return: dict
    """
    anns = load_json(path_ann)
    path, file_name = os.path.split(path_ann)
    base_path, _ = os.path.split(path)
    path = os.path.join(base_path, "images")
    anns["file_name"] = os.path.join(path, file_name[:-4] + "png")
    return anns


_NAME = "funsd"
_DESCRIPTION = (
    "FUNSD: Form Understanding in Noisy Scanned Documents. A dataset for Text Detection, Optical Character \n"
    " Recognition, Spatial Layout Analysis and Form Understanding."
)
_LICENSE = (
    "Use of the FUNSD Dataset is solely for non-commercial, research and educational purposes. The FUNSD  \n"
    "Dataset include annotations and images of real scanned forms. Licensee’s use of the images is governed \n"
    "by a copyright. Licensee is solely responsible for determining what additional licenses, clearances, \n"
    "consents and releases, if any, must be obtained for its use of the images. Original images are part of \n"
    "the dataset RVL-CDIP."
)

_URL = "https://guillaumejaume.github.io/FUNSD/download/"
_SPLITS: Mapping[str, str] = {"train": "training_data", "test": "testing_data"}
_TYPE = DatasetType.token_classification
_LOCATION = "funsd"
_ANNOTATION_FILES: Mapping[str, str] = {"train": "annotations", "test": "annotations"}

_INIT_CATEGORIES = [LayoutType.word, LayoutType.text]
_SUB_CATEGORIES: Dict[ObjectTypes, Dict[ObjectTypes, List[ObjectTypes]]]
_SUB_CATEGORIES = {
    LayoutType.word: {
        WordType.token_class: [TokenClasses.other, TokenClasses.question, TokenClasses.answer, TokenClasses.header],
        WordType.tag: [BioTag.inside, BioTag.outside, BioTag.begin],
        WordType.token_tag: [
            TokenClassWithTag.b_answer,
            TokenClassWithTag.b_header,
            TokenClassWithTag.b_question,
            TokenClassWithTag.i_answer,
            TokenClassWithTag.i_header,
            TokenClassWithTag.i_question,
            BioTag.outside,
        ],
    },
    LayoutType.text: {
        WordType.token_class: [TokenClasses.other, TokenClasses.question, TokenClasses.answer, TokenClasses.header]
    },
}


@dataset_registry.register("funsd")
class Funsd(_BuiltInDataset):
    """
    Funsd
    """

    _name = _NAME

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name=_NAME, description=_DESCRIPTION, license=_LICENSE, url=_URL, splits=_SPLITS, type=_TYPE)

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES, init_sub_categories=_SUB_CATEGORIES)

    def _builder(self) -> "FunsdBuilder":
        return FunsdBuilder(location=_LOCATION, annotation_files=_ANNOTATION_FILES)


class FunsdBuilder(DataFlowBaseBuilder):
    """
    Funsd dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the returns
        of the dataflow:

        `split:` Split of the dataset. "train" and "test" is available
        `load_image:` Will load the image for each datapoint.  Default: `False`
        `max_datapoints:` Will stop iterating after max_datapoints. Default: `None`

        :return: Dataflow
        """

        split = str(kwargs.get("split", "test"))
        load_image = kwargs.get("load_image", False)
        max_datapoints = kwargs.get("max_datapoints")

        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)

        # Load
        path_ann_files = self.get_workdir() / self.splits[split] / self.get_annotation_file(split)

        df = SerializerFiles.load(path_ann_files, ".json", max_datapoints)

        df = MapData(df, load_file)

        # Map
        categories_name_as_key = self.categories.get_categories(init=True, name_as_key=True)
        category_names_mapping = {
            "other": TokenClasses.other,
            "question": TokenClasses.question,
            "answer": TokenClasses.answer,
            "header": TokenClasses.header,
        }
        ner_token_to_id_mapping = self.categories.get_sub_categories(
            categories=LayoutType.word,
            sub_categories={LayoutType.word: [WordType.token_tag, WordType.tag, WordType.token_class]},
            keys=False,
            values_as_dict=True,
            name_as_key=True,
        )
        df = MapData(
            df,
            xfund_to_image(load_image, False, categories_name_as_key, category_names_mapping, ner_token_to_id_mapping),
        )
        if self.categories.is_filtered():
            df = MapData(
                df,
                filter_cat(  # pylint: disable=E1120
                    self.categories.get_categories(as_dict=False, filtered=True),
                    self.categories.get_categories(as_dict=False, filtered=False),
                ),
            )
        if self.categories.is_cat_to_sub_cat():
            df = MapData(
                df,
                cat_to_sub_cat(self.categories.get_categories(name_as_key=True), self.categories.cat_to_sub_cat),
            )
        return df



# -*- coding: utf-8 -*-
# File: test_deskew.py

# Copyright 2022 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module extern.deskew
"""

from numpy.testing import assert_array_equal
from pytest import mark

from deepdoctection.extern.deskew import Jdeskewer
from deepdoctection.utils.fs import load_image_from_file
from tests.test_utils import get_test_path


class TestJdeskewer:
    """
    Test Jdeskewer
    """

    @staticmethod
    @mark.requires_tf_or_pt
    def test_deskewer_transforms_image() -> None:
        """
        Detector deskews image and rotates it accordingly
        """

        # Arrange
        test_path_input_image = get_test_path() / "skewed_input.png"
        test_path_gt_image = get_test_path() / "skewed_gt.png"

        image = load_image_from_file(test_path_input_image)
        image_gt = load_image_from_file(test_path_gt_image)

        deskewer = Jdeskewer()

        # Act
        assert image is not None
        output_image = deskewer.transform(image)

        # Assert
        assert image_gt is not None
        assert_array_equal(image_gt, output_image)



# -*- coding: utf-8 -*-
# File: backbone.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")

"""
This file is modified from
<https://github.com/tensorpack/tensorpack/blob/master/examples/FasterRCNN/modeling/backbone.py>
"""

from contextlib import ExitStack, contextmanager

import numpy as np

# pylint: disable=import-error
import tensorflow as tf
from tensorpack import tfv1
from tensorpack.models import BatchNorm, Conv2D, MaxPooling, layer_register
from tensorpack.tfutils import argscope
from tensorpack.tfutils.varreplace import custom_getter_scope, freeze_variables

# pylint: enable=import-error


@layer_register(log_shape=True)
def GroupNorm(x, group=32, gamma_initializer=tf.constant_initializer(1.0)):
    """
    More code that reproduces the paper can be found at <https://github.com/ppwwyyxx/GroupNorm-reproduce/>.
    """
    shape = x.get_shape().as_list()
    ndims = len(shape)
    assert ndims == 4, shape
    chan = shape[1]
    assert chan % group == 0, chan
    group_size = chan // group

    orig_shape = tf.shape(x)
    h, w = orig_shape[2], orig_shape[3]

    x = tf.reshape(x, tf.stack([-1, group, group_size, h, w]))

    mean, var = tfv1.nn.moments(x, [2, 3, 4], keep_dims=True)

    new_shape = [1, group, group_size, 1, 1]

    beta = tfv1.get_variable("beta", [chan], initializer=tf.constant_initializer())
    beta = tf.reshape(beta, new_shape)

    gamma = tfv1.get_variable("gamma", [chan], initializer=gamma_initializer)
    gamma = tf.reshape(gamma, new_shape)

    out = tf.nn.batch_normalization(x, mean, var, beta, gamma, 1e-5, name="output")
    return tf.reshape(out, orig_shape, name="output")


def freeze_affine_getter(getter, *args, **kwargs):
    """
    Custom getter to freeze affine params inside bn
    """

    name = args[0] if len(args) else kwargs.get("name")  # pylint: disable=C1801
    if name.endswith("/gamma") or name.endswith("/beta"):
        kwargs["trainable"] = False
        ret = getter(*args, **kwargs)
        tf.add_to_collection(tf.GraphKeys.MODEL_VARIABLES, ret)
    else:
        ret = getter(*args, **kwargs)
    return ret


def maybe_reverse_pad(cfg, topleft, bottomright):
    """
    Returning chosen pad mode
    """
    if cfg.BACKBONE.TF_PAD_MODE:
        return [topleft, bottomright]
    return [bottomright, topleft]


@contextmanager
def backbone_scope(cfg, freeze):
    """
    Context scope for setting up the backbone

    :param cfg: The configuration instance as an AttrDict
    :param freeze: (bool) whether to freeze all the variables under the scope
    """

    def nonlin(x):
        x = get_norm(cfg)(x)
        return tf.nn.relu(x)

    with argscope([Conv2D, MaxPooling, BatchNorm], data_format="channels_first"), argscope(
        Conv2D,
        use_bias=False,
        activation=nonlin,
        kernel_initializer=tfv1.variance_scaling_initializer(scale=2.0, mode="fan_out"),
    ), ExitStack() as stack:
        if cfg.BACKBONE.NORM in ["FreezeBN", "SyncBN"]:
            if freeze or cfg.BACKBONE.NORM == "FreezeBN":
                stack.enter_context(argscope(BatchNorm, training=False))
            else:
                stack.enter_context(
                    argscope(BatchNorm, sync_statistics="nccl" if cfg.TRAINER == "replicated" else "horovod")
                )

        if freeze:
            stack.enter_context(freeze_variables(stop_gradient=False, skip_collection=True))
        else:
            # the layers are not completely freezed, but we may want to only freeze the affine
            if cfg.BACKBONE.FREEZE_AFFINE:
                stack.enter_context(custom_getter_scope(freeze_affine_getter))
        yield


def image_preprocess(image, cfg):
    """
    Preprocessing image by rescaling.

    :param image: tf.Tensor
    :param cfg: config
    :return: tf.Tensor
    """
    with tf.name_scope("image_preprocess"):
        if image.dtype.base_dtype != tf.float32:
            image = tf.cast(image, tf.float32)

        mean = cfg.PREPROC.PIXEL_MEAN
        std = np.asarray(cfg.PREPROC.PIXEL_STD)
        image_mean = tf.constant(mean, dtype=tf.float32)
        image_invstd = tf.constant(1.0 / std, dtype=tf.float32)
        image = (image - image_mean) * image_invstd
        return image


def get_norm(cfg, zero_init=False):
    """
    Return a norm with respect to config.

    :param cfg: config
    :param zero_init: choosing zero initializer or None
    :return: lambda func with norm applied to tf.Tensor
    """
    if cfg.BACKBONE.NORM == "None":
        return lambda x: x
    if cfg.BACKBONE.NORM == "GN":
        norm = GroupNorm
        layer_name = "gn"
    else:
        norm = BatchNorm
        layer_name = "bn"
    return lambda x: norm(layer_name, x, gamma_initializer=tf.zeros_initializer() if zero_init else None)


def resnet_shortcut(l, n_out, stride, activation=tf.identity):
    """
    Defining the skip connection in bottleneck

    :param l: tf.Tensor
    :param n_out: output dim
    :param stride: stride
    :param activation: An activation function
    :return: tf.Tensor
    """
    n_in = l.shape[1]
    if n_in != n_out:  # change dimension when channel is not the same
        return Conv2D("convshortcut", l, n_out, 1, strides=stride, activation=activation)
    return l


def resnet_bottleneck(l, ch_out, stride, cfg):
    """
    Defining the bottleneck for Resnet50/101 backbones

    :param l: tf.Tensor
    :param ch_out: number of output features
    :param stride: stride
    :param cfg: config
    :return: tf.Tensor
    """
    shortcut = l

    l = Conv2D("conv1", l, ch_out, 1, strides=1)
    if stride == 2:
        l = tf.pad(l, [[0, 0], [0, 0], maybe_reverse_pad(cfg, 0, 1), maybe_reverse_pad(cfg, 0, 1)])
        l = Conv2D("conv2", l, ch_out, 3, strides=2, padding="VALID")
    else:
        l = Conv2D("conv2", l, ch_out, 3, strides=stride)
    if cfg.BACKBONE.NORM != "None":
        l = Conv2D("conv3", l, ch_out * 4, 1, activation=get_norm(cfg, zero_init=True))
    else:
        l = Conv2D("conv3", l, ch_out * 4, 1, activation=tf.identity, kernel_initializer=tf.constant_initializer())
    ret = l + resnet_shortcut(shortcut, ch_out * 4, stride, activation=get_norm(cfg))
    return tf.nn.relu(ret, name="output")


def resnext32x4d_bottleneck(l, ch_out, stride, cfg):
    """
    Defining Resnext bottleneck <https://arxiv.org/abs/1611.05431>

    :param l: tf.Tensor
    :param ch_out: number of output features
    :param stride: stride
    :param cfg: config
    :return: tf.Tensor
    """
    shortcut = l
    l = Conv2D("conv1", l, ch_out * 2, 1, stride=1)
    l = Conv2D("conv2", l, ch_out * 2, 3, stride=stride, split=32)
    l = Conv2D("conv3", l, ch_out * 4, 1, activation=get_norm(cfg, zero_init=True))
    ret = l + resnet_shortcut(shortcut, ch_out * 4, stride, activation=get_norm(cfg))
    return tf.nn.relu(ret, name="output")


def resnet_group(name, l, block_func, features, count, stride, cfg):
    """
    Defining resnet groups

    :param name: name of group
    :param l: tf.Tensor
    :param block_func: the bottleneck function
    :param features: number of features
    :param count: number of successive bottleneck function
    :param stride: stride
    :param cfg: config
    :return: tf.Tensor
    """
    with tfv1.variable_scope(name):
        for i in range(0, count):
            with tfv1.variable_scope(f"block{i}"):
                l = block_func(l, features, stride if i == 0 else 1, cfg)
    return l


def resnet_fpn_backbone(image, cfg):
    """
    Setup of the full FPN backbone.

    :param image: tf.Tensor
    :param cfg: config
    :return: tf.Tensor for level c2,c3,c4,c5
    """
    num_blocks = cfg.BACKBONE.RESNET_NUM_BLOCKS
    freeze_at = cfg.BACKBONE.FREEZE_AT
    shape2d = tf.shape(image)[2:]
    mult = float(cfg.FPN.RESOLUTION_REQUIREMENT)
    new_shape2d = tf.cast(tf.math.ceil(tf.cast(shape2d, tf.float32) / mult) * mult, tf.int32)
    pad_shape2d = new_shape2d - shape2d
    assert len(num_blocks) == 4, num_blocks
    with backbone_scope(cfg, freeze=freeze_at > 0):
        chan = image.shape[1]
        pad_base = maybe_reverse_pad(cfg, 2, 3)
        l = tf.pad(
            image,
            tf.stack(
                [
                    [0, 0],
                    [0, 0],
                    [pad_base[0], pad_base[1] + pad_shape2d[0]],
                    [pad_base[0], pad_base[1] + pad_shape2d[1]],
                ]
            ),
        )
        l.set_shape([None, chan, None, None])
        l = Conv2D("conv0", l, 64, 7, strides=2, padding="VALID")
        l = tf.pad(l, [[0, 0], [0, 0], maybe_reverse_pad(cfg, 0, 1), maybe_reverse_pad(cfg, 0, 1)])
        l = MaxPooling("pool0", l, 3, strides=2, padding="VALID")

    bottleneck = resnet_bottleneck if cfg.BACKBONE.BOTTLENECK == "resnet" else resnext32x4d_bottleneck
    with backbone_scope(cfg=cfg, freeze=freeze_at > 1):
        c2 = resnet_group("group0", l, bottleneck, 64, num_blocks[0], 1, cfg)
    with backbone_scope(cfg=cfg, freeze=False):
        c3 = resnet_group("group1", c2, bottleneck, 128, num_blocks[1], 2, cfg)
        c4 = resnet_group("group2", c3, bottleneck, 256, num_blocks[2], 2, cfg)
        c5 = resnet_group("group3", c4, bottleneck, 512, num_blocks[3], 2, cfg)

    # 32x downsampling up to now
    # size of c5: ceil(input/32)
    return c2, c3, c4, c5



# -*- coding: utf-8 -*-
# File: __init__.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This package everything that is related to the **core data model**.

A **fundamental architectural decision** in the deep doctection framework is to convert all data, be it those from
datasets or the predictions from the pipelines, into a **standardized format**. This procedure leads to the following
simplifications:

- Data of the training environment can be transported in the production environment (i.e. through the pipelines)
  without further adjustments.

- Datasets of different origins can be merged quickly, so that training data with greater variability arise.

- Pipeline environment components can be executed one after the other without conversion measures.

The disadvantage of carrying out any redundant transformations and thus experiencing a loss of performance is accepted.
After all, the point here is not to provide an optimal processing environment.
"""

from .annotation import *
from .box import *
from .convert import *
from .image import Image
from .view import *



# -*- coding: utf-8 -*-
# File: systools.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
module for various basic functions that are needed everywhere
"""

import os
from pathlib import Path

from .develop import deprecated
from .settings import CONFIGS, DATASET_DIR, MODEL_DIR, PATH

__all__ = ["sub_path", "get_package_path", "get_configs_dir_path", "get_weights_dir_path", "get_dataset_dir_path"]


def get_package_path() -> Path:
    """
    :return: full base path of this package
    """
    return PATH


def get_weights_dir_path() -> Path:
    """
    :return: full base path to the model dir
    """
    return MODEL_DIR


def get_configs_dir_path() -> Path:
    """
    :return: full base path to the configs dir
    """
    return CONFIGS


def get_dataset_dir_path() -> Path:
    """
    :return: full base path to the dataset dir
    """
    return DATASET_DIR


@deprecated("Use pathlib operations instead", "2022-06-08")
def sub_path(anchor_dir: str, *paths: str) -> str:
    """
    Generate a path from the anchor directory and various paths args.

        sub_path(/path/to,"dir1","dir2")

    will return `/path/to/dir1/dir2`

    :param anchor_dir: anchor directory
    :param paths: args of directories that should be added to path
    :return: sub_path
    """
    return os.path.join(os.path.dirname(os.path.abspath(anchor_dir)), *paths)
